#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ##########################
# ADAPT copyright statement:
# Copyright 2006 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software. This software is distributed under the GNU Lesser General Public License version 3 or later.
# ##########################
import sys
import traceback
import time
from misc_const import SqlConnectMode, sql_connect_mode, GetDatabaseAccessLock

try:
    import mysql
    import mysql.connector
    sql_connect_mode = SqlConnectMode.mysql
except ImportError:
    print('ADAPT requires the MySQL python connector maintained by Oracle at https://dev.mysql.com/downloads/connector/python/')
    sys.exit(1)


class DatabaseConnection():
    LastSqlStatementWas = None
    
    def __init__(self, database_hostname, database_port, database_user, database_password, database_name):
        # thread locking handled in _db_get_handles
        self._db_get_handles(database_hostname, database_port, database_user, database_password, database_name)
    
    def __del__(self):
        try:
            while not GetDatabaseAccessLock().acquire():
                print("Could not acquire lock in DatabaseConnection(). Sleeping for one second.")
                time.sleep(1)
            try:
                self._close()
            finally:
                GetDatabaseAccessLock().release()
        except Exception:
            pass
            
    def ExplicitlyClose(self):
        self._close()
        
    def _close(self):
    
        self._db_conn.commit()
        self._db_cursor.close()
        del self._db_cursor
        self._db_readcursor.close()
        del self._db_readcursor
        self._db_conn.close()
            
    def ExecuteConnectionCommit(self):
        while not GetDatabaseAccessLock().acquire():
            print("Could not acquire lock in ExecuteConnectionCommit. Sleeping for one second.")
            time.sleep(1)
        try:
            self._db_conn.commit()
        finally:
            GetDatabaseAccessLock().release()
    
    def _db_get_handles(self, database_hostname, database_port, database_user, database_password, database_name):
        while not GetDatabaseAccessLock().acquire():
            print("Could not acquire lock in _db_get_handles. Sleeping for one second.")
            time.sleep(1)
            
        try:
            self._db_init(database_hostname, database_port, database_user, database_password, database_name)
        finally:
            GetDatabaseAccessLock().release()
    
    def ExecuteReadQuery(self, SqlQuery):
        while not GetDatabaseAccessLock().acquire():
            print("Could not acquire lock in ExecuteReadQuery. Sleeping for one second.")
            time.sleep(1)
        
        Result = None
        
        try:
            LastSqlStatementWas = SqlQuery
            #self._db_conn.commit()
            self._db_readcursor.execute(SqlQuery)
            Result = self._db_readcursor.fetchall()
        except mysql.connector.InterfaceError:
            raise RuntimeError("Database interface error in ExecuteReadQuery.")
        except mysql.connector.OperationalError:
            raise RuntimeError("Database interface error in ExecuteReadQuery.")
        except mysql.connector.errors.ProgrammingError:
            raise RuntimeError("Error somewhere in SQL statement. This is sometimes due to a string with unescaped special characters. Text of statement:\n %s" % SqlQuery)
        finally:
            GetDatabaseAccessLock().release()
        
        return Result
        
    def ExecuteLegacyReadQuery(self, SqlQuery):
        while not GetDatabaseAccessLock().acquire():
            print("Could not acquire lock in ExecuteLegacyReadQuery. Sleeping for one second.")
            time.sleep(1)
        
        Result = None
        
        try:
            LastSqlStatementWas = SqlQuery
            self._db_conn.commit()
            self._db_cursor.execute(SqlQuery)
            Result = self._db_cursor.fetchall()
        except mysql.connector.InterfaceError:
            raise RuntimeError("Database interface error in ExecuteLegacyReadQuery.")
        except mysql.connector.OperationalError:
            raise RuntimeError("Database interface error in ExecuteLegacyReadQuery.")
        except mysql.connector.errors.ProgrammingError:
            raise RuntimeError("Error somewhere in SQL statement. This is sometimes due to a string with unescaped special characters. Text of statement:\n %s" % SqlQuery)
        finally:
            GetDatabaseAccessLock().release()
        
        return Result
    
    def ExecuteWriteQuery(self, SqlQuery, Val=None):
        while not GetDatabaseAccessLock().acquire():
            print("Could not acquire lock in ExecuteWriteQuery. Sleeping for one second.")
            time.sleep(1)
        try:
            if Val is None:
                LastSqlStatementWas = SqlQuery
                self._db_conn.commit()
                self._db_cursor.execute(SqlQuery)
            else:
                LastSqlStatementWas = SqlQuery, Val
                self._db_conn.commit()
                self._db_cursor.execute(SqlQuery, Val)
        except mysql.connector.InterfaceError:
            raise RuntimeError("Database interface error in ExecuteWriteQuery.")
        except mysql.connector.OperationalError:
            raise RuntimeError("Database interface error in ExecuteWriteQuery.")
        except mysql.connector.errors.ProgrammingError:
            raise RuntimeError("Error somewhere in SQL statement. This is sometimes due to a string with unescaped special characters. Text of statement:\n %s" % SqlQuery)
        finally:
            GetDatabaseAccessLock().release()
            
    def _db_init(self, h, po, u, pa, d):
        try:
            connected = 0
            for i in range(5):
                connected = 0
                try:
                    if (sql_connect_mode == SqlConnectMode.mysql):
                        tries = 0
                        while tries < 5:
                            try:
                                self._db_conn = mysql.connector.connect(host=h, port=po, user=u, password=pa, database=d, connection_timeout=86400)
                                tries = 5
                            except:
                                print('%s: Failed to connect to the database. Retrying.' % (str(time.asctime())))
                                tries += 1
                                time.sleep(10)
                                if tries == 5:
                                    raise RuntimeError("%s: DatabaseConnection() failed five times. Giving up." % (str(time.asctime())))
                                    sys.exit(1)
                    else:
                        raise RuntimeError("Unknown sql_connect_mode value in DatabaseConnection().")
                    
                    # Write cursor (can read also, but result set does not contain field names).
                    self._db_cursor = self._db_conn.cursor()
                    self._db_cursor.execute('SET AUTOCOMMIT=0')
                    self._db_cursor.execute('SET SESSION wait_timeout = 3600 * 2400')
                    
                    if (sql_connect_mode == SqlConnectMode.mysql):
                        self._db_readcursor = self._db_conn.cursor(dictionary=True)
                    else:
                        raise RuntimeError("Unknown sql_connect_mode value in DatabaseConnection().")
                        
                    self._db_readcursor.execute('SET AUTOCOMMIT=0')
                    self._db_readcursor.execute('SET SESSION wait_timeout = 3600 * 2400')
                    
                    connected = 1
                except Exception:
                    traceback.print_exc()
                
                if connected:
                    break
                else:
                    sleeptime = 20
                    print("Couldn't connect to database, retrying in %s seconds..." % str(sleeptime))
                    time.sleep(sleeptime)
            if not connected:
                raise RuntimeError('ERROR: Tried and failed 5 times to connect to database.')
        except Exception:
            sys.stderr.write("\n")
            traceback.print_exc()
            sys.stderr.write("ERROR: Unspecified error communicating with database.\n")
            sys.exit(1)
