#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ##########################
# ADAPT copyright statement:
# Copyright 2006 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software. This software is distributed under the GNU Lesser General Public License version 3 or later.
# ##########################
import os
import sys
import re
import time


if int(sys.version[0]) == 2:
    import ConfigParser
elif int(sys.version[0]) > 2:
    import configparser
else:
    raise ValueError(str(time.asctime()) + ': Unknown python version in adapt_config.py.')


class AdaptConfig:
        
    def __init__():
        pass

    @staticmethod
    def get_config_value(section, option):
        if int(sys.version[0]) == 2:
            p = ConfigParser.ConfigParser()
        elif int(sys.version[0]) > 2:
            p = configparser.ConfigParser()
        else:
            raise ValueError(str(time.asctime()) + ': Unknown python version in adapt_config.py.')
        p.read(AdaptConfig.get_rc_filename())
        if not p.has_section(section):
            return None
        if not p.has_option(section, option):
            return None
        out = p.get(section, option)
        if out == '':
            out = None
        if '~' in out:
            out = os.path.expanduser(out)
        return out
        
    @staticmethod
    def get_server_jobmanager():
        return AdaptConfig.get_config_value('server', 'server_jobmanager')

    @staticmethod
    def get_server_log():
        return AdaptConfig.get_config_value('server', 'server_log')

    @staticmethod
    def get_server_logwebmin():
        return AdaptConfig.get_config_value('server', 'server_logwebmin')

    @staticmethod
    def get_server_webmin():
        v = AdaptConfig.get_config_value('server', 'server_webmin')
        if not v:
            raise Exception(str(time.asctime()) + ': server/webmin undefined in %s.' % (AdaptConfig.get_rc_filename()))
        else:
            assert v.find(' ') == -1
            return v
            
    @staticmethod
    def get_server_adaptemp():
        v = AdaptConfig.get_config_value('server', 'server_adaptemp')
        if not v:
            raise Exception(str(time.asctime()) + ': server/adaptemp undefined in %s.' % (AdaptConfig.get_rc_filename()))
        else:
            assert v.find(' ') == -1
            return v
            
    @staticmethod
    def get_webmin_users():
        return AdaptConfig.get_config_value('webmin', 'webmin_users')
        
    @staticmethod
    def get_webmin_port():
        v = AdaptConfig.get_config_value('webmin', 'webmin_port')
        if not v:
            v = '47833'
        return int(v)
        
    @staticmethod
    def get_webmin_baseurl():
        v = AdaptConfig.get_config_value('webmin', 'webmin_baseurl')
        if not v:
            v = ''
        return v
        
    @staticmethod
    def get_database_hostname():
        return AdaptConfig.get_config_value('database', 'database_host') or AdaptConfig.get_config_value('database', 'hostname')
        
    @staticmethod
    def get_database_port():
        v = AdaptConfig.get_config_value('database', 'database_port')
        return v and int(v) or None
        
    @staticmethod
    def get_database_user():
        result = AdaptConfig.get_config_value('database', 'database_user')
        if(result is None):
            raise Exception(str(time.asctime()) + ': Database user not set in %s.' % (AdaptConfig.get_rc_filename()))
        return result
        
    @staticmethod
    def get_database_pass():
        return AdaptConfig.get_config_value('database', 'database_pass')
        
    @staticmethod
    def get_database_db():
        return AdaptConfig.get_config_value('database', 'database_db')
        
    @staticmethod
    def get_rc_filename():
        result = None
    
        if os.getenv('NCENGINE_RCFILE'):
            result = os.getenv('NCENGINE_RCFILE')
        else:
            ADAPTRC = os.getenv('ADAPTRC')
            if ADAPTRC is None:
                raise Exception(str(time.asctime()) + ": ADAPTRC environment variable not set.")
            result = os.path.join(ADAPTRC, '.adaptrc')

        return result

    @staticmethod
    def get_script_loc():
        return AdaptConfig.get_config_value('script', 'scriptloc')

    @staticmethod
    def get_rove_hostname():
        return AdaptConfig.get_config_value('rove', 'host')

    @staticmethod
    def get_rove_port():
        v = AdaptConfig.get_config_value('rove', 'port')
        return v and int(v) or None

    @staticmethod
    def get_ssh_ssh():
        v = AdaptConfig.get_config_value('ssh', 'ssh_ssh')
        if not v:
            raise Exception(str(time.asctime()) + ': ssh/ssh undefined in %s.' % (AdaptConfig.get_rc_filename()))
        return v

    @staticmethod
    def get_ssh_scp():
        v = AdaptConfig.get_config_value('ssh', 'ssh_scp')
        if not v:
            raise Exception(str(time.asctime()) + ': ssh/scp undefined in %s.' % (AdaptConfig.get_rc_filename()))
        return v

    @staticmethod
    def get_ssh_shared_dirs():
        v = AdaptConfig.get_config_value('ssh', 'ssh_shared_dirs')
        if not v:
            v = False
        else:
            if (v.lower() == 'true'):
                v = True
            elif (v.lower() == 'false'):
                v = False
            else:
                print('''Expected value of 'true' or 'false' for ssh_shared_dirs. Assuming false.''')
                v = False
        return v

    @staticmethod
    def get_ssh_hostlist():
        v = AdaptConfig.get_config_value('ssh', 'ssh_hosts')
        if not v:
            raise Exception(str(time.asctime()) + ': ssh/hosts undefined in %s.' % (AdaptConfig.get_rc_filename()))
        return v.split()

    @staticmethod
    def get_ssh_directory(host):
        v = AdaptConfig.get_config_value('ssh', 'ssh_directory')
        if not v:
            raise Exception(str(time.asctime()) + ': ssh/directory undefined in %s.' % (AdaptConfig.get_rc_filename()))
        
        if v.find(':') != -1:
            exprs = v.split()
            for expr in exprs:
                regexp, dir = expr.split(':')
                if '.' in regexp or '^' in regexp or '$' in regexp or '*' in regexp or '+' in regexp or '?' in regexp or '{' in regexp or '}' in regexp or '[' in regexp or ']' in regexp or '\\' in regexp or '|' in regexp or '(' in regexp or ')' in regexp:
                    # Only go down this road if we have a regexp metacharacter.
                    if re.search(regexp, host):
                        if '~' in dir:
                            dir = os.path.expanduser(dir)
                        return dir
                else:
                    if regexp == host:
                        if '~' in dir:
                            dir = os.path.expanduser(dir)
                        return dir
            sys.stderr.write(str(time.asctime()) + ": ERROR: host %s doesn't match any regular expressions in ssh directory list %s (check config file)\n" % (host, v))
            sys.exit(1)
        else:
            print('Expected form for entries in ssh_directory is: host1:directory1 host2:directory2')
            sys.exit(1)
    
    @staticmethod
    def get_ssh_job_launch_limit():
        v = AdaptConfig.get_config_value('ssh', 'ssh_job_launch_limit')
        if not v:
            v = '200'
        return int(v)
    
    @staticmethod
    def get_ssh_2stepcopy():
        v = AdaptConfig.get_config_value('ssh', 'ssh_2stepcopy')
        if not v:
            return False
        if v == '1' or v.lower() == 'true' or v.lower() == 'y' or v.lower() == 'yes':
            return True
        else:
            return False
