#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ##########################
# ADAPT copyright statement:
# Copyright 2006 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software. This software is distributed under the GNU Lesser General Public License version 3 or later.
# ##########################
# ADAPT module description for adaptlibs.py: This module contains many commonly-used functions for ADAPT.
from __future__ import print_function
from __future__ import division
from adapt_config import AdaptConfig
from AdaptDatabaseIO import DatabaseCalls
import adaptlibs
import os
import sys
import re
import pickle
import json
import tempfile
import shutil
import time
import zipfile
import threading
from multiprocessing.dummy import Pool as ThreadPool
import subprocess
import operator
import random


def GetDataAccessObject():
    
    return DatabaseCalls.GetOrCreateInstance(AdaptConfig.get_database_hostname(), AdaptConfig.get_database_port(), AdaptConfig.get_database_user(), AdaptConfig.get_database_pass(), AdaptConfig.get_database_db())


def CloneDataAccessObject():
    return GetDataAccessObject().Clone()


def get_platform():
    result = None
    
    if sys.platform == "linux" or sys.platform == "linux2":
        result = "linux"
    elif sys.platform == "darwin":
        result = "osx"
    elif sys.platform == "win32":
        result = "windows"
    elif sys.platform == "windows":
        result = "windows"
    elif sys.platform == "cygwin":
        result = "cygwin"
    else:
        raise Exception(sys.platform + " unknown platform.")
    
    return result


def err(message):
    sys.stderr.write("ERROR: %s\n" % (message))
    sys.exit(1)


def which(filename):
    import stat
    if 'PATH' not in os.environ or os.environ['PATH'] == '':
        p = os.defpath
    else:
        p = os.environ['PATH']
    pathlist = p.split(os.pathsep)
    for path in pathlist:
        f = os.path.join(path, filename)
        if os.access(f, os.X_OK) and not stat.S_ISDIR(os.stat(f)[stat.ST_MODE]):
            return f
    return None


def requireutil(executables):
    for exe in executables:
        if not which(exe):
            sys.stderr.write("ERROR: cannot find required util %s\n" % (exe))
            sys.exit(1)


def unixfind(path):
    out = []
    for (pth, dames, fnames) in os.walk(path):
        for fn in fnames:
            out.append(os.path.join(pth, fn))
    out.sort()
    return out

__safe_chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"


def shell_quote(s):
    """Quote s so that it is safe for all common shells.
    """
    res = []
    for c in s:
        if c in __safe_chars:
            res.append(c)
        else:
            res.append("=%02x" % ord(c))
    return ''.join(res)


def shell_unquote(s):
    """Unquote a string quoted by shell_quote.
    """
    if s == "":
        return ""
    
    frags = s.split("=")
    res = [frags[0]]
    for f in frags[1:]:
        res.append(chr(int(f[:2], 0x10)))
        res.append(f[2:])
    return ''.join(res)


def shell_escape(s):
    res = []
    for c in s:
        if c not in __safe_chars:
            res.append('\\')
        res.append(c)
    return ''.join(res)


def parsetime(timespec):
    if re.search("^[0-9]+$", timespec):
        return int(timespec)
    elif re.search("^[0-9]+s$", timespec):
        return int(timespec[0:len(timespec) - 1])
    elif re.search("^[0-9]+m$", timespec):
        return int(timespec[0:len(timespec) - 1]) * 60
    elif re.search("^[0-9]+h$", timespec):
        return int(timespec[0:len(timespec) - 1]) * 3600
    elif re.search("^[0-9]+d$", timespec):
        return int(timespec[0:len(timespec) - 1]) * 3600 * 24
    else:
        raise RuntimeError('ERROR: parsing %s' % (timespec))


def get_checkpoint_script_name():
    return '.adapt-checkpoint-script'


def get_checkpoint_taken_name():
    return '.adapt-checkpoint-taken'


# Check the [1min, 5min, 15min] system load on the named host. This will not work for a Windows host.
def load_check(host):
    ssh = AdaptConfig.get_ssh_ssh()
    cmd = ssh + ' ' + host + ' uptime'
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    stdout, stderr = p.communicate()
    
    if int(sys.version[0] == 2):
        try:
            load = stdout.split('load average: ')[1].split()
            for x in range(0, len(load)):
                load[x] = load[x].replace(',', '')
        except:
            print('Communication error with host %s.' % host)
            load = False
    else:
        try:
            load = stdout.split(b'load average: ')[1].split()
            for x in range(0, len(load)):
                load[x] = load[x].replace(b',', b'')
        except:
            print('Communication error with host %s.' % host)
            load = False
    
    return load


# Check the [1min, 5min, 15min] system load on all hosts in .adaptrc. This will not work for a Windows host.
def load_check_all():
    hostlist = AdaptConfig.get_ssh_hostlist()
    unique_hosts = sorted(list(set(hostlist)))
    pool = ThreadPool(len(unique_hosts))
    load = pool.map(adaptlibs.load_check, unique_hosts)
    
    return (unique_hosts, load)


# Check the number of ADAPT jobs running on each host.
def adapt_jobs_check():
    all_jobs_running = adaptlibs.GetDataAccessObject().jobsSelectAllWhereState(state='R')
    host_jobs = {}
    for job in all_jobs_running:
        # Add the host to the dictionary if it isn't already there.
        if job[3] not in host_jobs:
            host_jobs[job[3]] = 0
        host_jobs[job[3]] += 1
    
    return host_jobs


# Find the number of free slots on each host using potentially two methods.
def free_adapt_slots(method='adapt_jobs_check'):
    # Method refers to using adapt_jobs_check, load_check, or both for load checking.
    if method not in ['adapt_jobs_check', 'load_check', 'both']:
        print("Input error. Acceptable values for method are ['adapt_jobs_check', 'load_check', 'both']")
    
    # How many total slots per host?
    hostlist = AdaptConfig.get_ssh_hostlist()
    unique_hosts = list(set(hostlist))
    host_slots = {}
    for host in unique_hosts:
        host_slots[host] = hostlist.count(host)
    
    free_slots = {}
    if method == 'adapt_jobs_check':
        # Check host load using only the ADAPT database. This may neglect other processes on the host.
        host_jobs = adaptlibs.adapt_jobs_check()
        for host in host_slots:
            try:
                free_slots[host] = host_slots[host] - host_jobs[host]
            except KeyError:
                # No jobs running on host according to the database, indicates all are free.
                free_slots[host] = host_slots[host]
    elif method == 'load_check':
        # Check host load using only the system load. This may neglect ADAPT jobs that are just starting.
        host_load = {}
        for host in host_slots:
            try:
                host_load[host] = int(round(float(adaptlibs.load_check(host)[0])))
                free_slots[host] = host_slots[host] - host_load[host]
            except TypeError:
                # load_check will return False if a communication error occurs. We cannot evaluate this host's load with this method.
                host_load[host] = False
    elif method == 'both':
        # Check host load using a combination of the above methods.
        
        host_jobs = adaptlibs.adapt_jobs_check()
        
        host_load = {}
        for host in host_slots:
            try:
                host_load[host] = int(round(float(adaptlibs.load_check(host)[0])))
            except TypeError:
                # load_check will return False if a communication error occurs. We cannot evaluate this host's load with this method.
                host_load[host] = False
        
        # First subtract based on jobs in the database.
        for host in host_slots:
            try:
                free_slots[host] = host_slots[host] - host_jobs[host]
            except KeyError:
                # No jobs running on host according to the database, indicates all are free.
                free_slots[host] = host_slots[host]
        
        # Next subtract based on system load. If there is no system load available, assume there are 0 slots available.
        host_load = {}
        for host in host_slots:
            try:
                host_load[host] = int(round(float(adaptlibs.load_check(host)[0])))
                free_slots[host] = free_slots[host] - host_load[host]
            except TypeError:
                # load_check will return False if a communication error occurs. We cannot evaluate this host's load with this method.
                host_load[host] = False
                free_slots[host] = 0
    
    return free_slots


def plot_exp_progress(expid):
    # Plot the progress of an experiment. Number of jobs running and cumulative jobs finished against time.
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import matplotlib.ticker as mtick
    import numpy as np
    # This uses numpy for speed.
    
    all_exp_jobs = adaptlibs.GetDataAccessObject().jobsSelectAllWhereInExpID(expid=expid)
    plusses = [x[5] for x in all_exp_jobs if x[5] is not None]
    minuses = [x[6] for x in all_exp_jobs if x[6] is not None]
    overall = plusses + minuses
    increments = [1 for x in plusses] + [-1 for x in minuses]
    overall, increments = zip(*sorted(zip(overall, increments)))
    running_jobs = list(np.cumsum(increments))
    
    completed_job_times = [x[6] for x in all_exp_jobs if x[6] is not None and x[2] == 'C']
    completed_job_times = sorted(completed_job_times)
    job_finished_plusses = [1] * len(completed_job_times)
    cumulative_jobs_finished = list(np.cumsum(job_finished_plusses))
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.semilogy(overall, running_jobs, label='Currently Running', color='blue')
    ax.semilogy(completed_job_times, cumulative_jobs_finished, label='Successfully Finished', color='black')
    
    plt.xticks(rotation=70)
    plt.xlabel('Date and Time')
    plt.ylabel('Number of ADAPT Jobs')
    ax.grid(b=True, which='major', color='gray', linestyle='-')
    ax.grid(b=True, which='minor', color='gray', linestyle='-')
    plt.legend(loc='best')
    
    thisdir = os.getcwd()
    pdf_file = os.path.join(thisdir, 'ADAPT-job-progress-%s.pdf' % str(expid))
    png_file = os.path.join(thisdir, 'ADAPT-job-progress-%s.png' % str(expid))
    
    plt.savefig(pdf_file, bbox_inches='tight')
    plt.savefig(png_file, bbox_inches='tight')
    plt.close()
    return pdf_file, png_file


def find_files(dir):
    out = []
    if not os.path.exists(dir):
        sys.stderr.write("ERROR: %s does not exist\n" % (dir))
        sys.exit(1)
    for root, dirs, files in os.walk(dir, topdown=False):
        for name in files:
            fn = os.path.join(root, name)
            out.append(fn)
    return out


def get_adapt_d_dir():
    d = AdaptConfig.get_server_webmin()
    #print("adapt_d_dir is " + str(d))
    
    if not os.access(d, os.F_OK):
        print ("Cannot access " + d + ". Creating folder...")
        os.mkdir(d)
        print ("Folder created.")
    return d


def checkpoint_branch(bid, jid, checkpointscript, hostname, rundir, pid, wait_completion):
    ssh = AdaptConfig.get_ssh_ssh()
    scp = AdaptConfig.get_ssh_scp()
    print('checkpointing branch', bid, 'jobid', jid, 'pid', pid)
    (tempfd, tempfn) = tempfile.mkstemp(suffix='.scr', dir=AdaptConfig.get_server_adaptemp())
    os.close(tempfd)
    open(tempfn, 'wb').write(checkpointscript)
    script_name = adaptlibs.get_checkpoint_script_name()
    print("scp=" + scp)
    
    cmd = '%s %s %s:%s' % (scp, tempfn, hostname, os.path.join(rundir, script_name))
    
    command_done = False
    tries = 0
    while (command_done is False) and (tries < 5):
        try:
            rc = subprocess.call(cmd, shell=True)
            tries += 1
            command_done = True
        except:
            tries += 1
            print('Error running %s. Retrying.' % (cmd))
            #sys.stderr.write("ERROR: running %s\n" % (cmd))
            #raise RuntimeError('ERROR: running %s' % (cmd))
    if (command_done is False) and (tries >= 5):
        print('Tried and failed 5 times to run %s.' % (cmd))
        raise Exception('Command %s failed for unspecified reason with %s result.' % (cmd, str(rc)))
    
    os.remove(tempfn)
    
    cmd = '%s %s "cd %s && chmod 755 %s && ./%s"' % (ssh, hostname, rundir, script_name, script_name)

    command_done = False
    tries = 0
    while (command_done is False) and (tries < 5):
        try:
            rc = subprocess.call(cmd, shell=True)
            tries += 1
            command_done = True
        except:
            tries += 1
            print('Error running %s. Retrying.' % (cmd))
            #sys.stderr.write("ERROR: running %s\n" % (cmd))
            #raise RuntimeError('ERROR: running %s' % (cmd))
    if (command_done is False) and (tries >= 5):
        print('Tried and failed 5 times to run %s.' % (cmd))
        raise Exception('Command %s failed for unspecified reason with %s result.' % (cmd, str(rc)))
    
    elif wait_completion:
        for seconds in range(10):
            cmd = '%s %s kill -0 %d 2>/dev/null' % (ssh, hostname, int(pid))
            print('+%s' % (cmd))
            rc = subprocess.call(cmd, shell=True)
            if rc == 0:
                time.sleep(seconds)
            else:
                break
        
        while True:
            secs = 1
            rows = adaptlibs.GetDataAccessObject().jobsSelectAllWhereStateAndJobID(jobid=jid, state='R')
            if len(rows) > 0:
                print('jobid', jid, 'not committed yet')
                time.sleep(secs)
                secs += 1
            else:
                break


def kill_remote_pid_group(hostname, pid, wait_completion):
    ssh = AdaptConfig.get_ssh_ssh()
    cmd = '%s %s kill -- -%d' % (ssh, hostname, int(pid))
    print(cmd)
    rc = subprocess.call(cmd, shell=True)
    if rc != 0:
        sys.stderr.write("WARNING:  running command %s\n" % (cmd))
    elif wait_completion:
        for seconds in range(10):
            cmd = '%s %s kill -0 %d 2>/dev/null' % (ssh, hostname, int(pid))
            print('+%s' % (cmd))
            rc = subprocess.call(cmd, shell=True)
            if rc == 0:
                time.sleep(seconds)
            else:
                break


def checkpoint_or_kill_branch(bid, wait_completion):
    while True:
        rows = adaptlibs.GetDataAccessObject().jobs_branches_experimentsSelectRunningJobsWhereBranchID(branchid=bid)
        if len(rows) == 0:
            print('no such branch', bid, 'cannot checkpoint or kill it!')
            return
        scr, jid, rundir, hostname, pid = rows[0]
        if pid is None:
            print('pid for bid %s is still null, waiting...' % (bid))
            time.sleep(2)
        else:
            break
    if scr:
        checkpoint_branch(bid, jid, scr, hostname, rundir, pid, wait_completion)
    else:
        kill_remote_pid_group(hostname, pid, wait_completion)
        adaptlibs.GetDataAccessObject().jobsUpdateStatePIDWhereID(state='A', branchid=bid)


def kill_active_jobs(eid=None):
    
    reps = 0
    while True:
        if eid:
            rows = adaptlibs.GetDataAccessObject().jobs_branches_experimentsSelectRunningJobsWhereExpID(expid=eid)
        else:
            rows = adaptlibs.GetDataAccessObject().jobs_branches_experimentsSelectRunningJobsAll()
        pending = 0
        for row in rows:
            checkpointscript, branchid, jobid, rundir, hostname, pid = row
            if not pid:
                pending = 1
        if pending:
            reps += 1
            if reps == 2:
                print('Giving up on waiting for jobs to start up.')
                break
            print('Some jobs are still pending start up. Waiting 10 seconds...')
            time.sleep(10.0)
        else:
            break
    
    jobids_aborted = []
    jobids_checkpointed = []
    for row in rows:
        checkpointscript, branchid, jobid, rundir, hostname, pid = row
        if len(checkpointscript) > 0:
            adaptlibs.checkpoint_branch(branchid, jobid, checkpointscript, hostname, rundir, pid, False)
            jobids_checkpointed.append(jobid)
        elif pid:
            adaptlibs.kill_remote_pid_group(hostname, pid, True)
            jobids_aborted.append(jobid)
    if len(jobids_aborted) > 0:
        adaptlibs.GetDataAccessObject().jobsUpdateStatePIDWhereinID(jobids_aborted=jobids_aborted)
    
    b4 = time.time()
    for jobid in jobids_checkpointed:
        ckpoint_complete = 0
        while not ckpoint_complete:
            rows = adaptlibs.GetDataAccessObject().jobsSelectAllWhereStateAndJobID(jobid=jobid, state='R')
            n = len(rows)
            
            if n == 0:
                ckpoint_complete = 1
            else:
                print('Waiting for checkpoint to complete for job', jobid)
                time.sleep(30)
                now = time.time()
                elapsed = now - b4
                if elapsed > (60 * 20):
                    sys.stderr.write("ERROR: took too long (20 minutes) to checkpoint, please manually cleanup jobs\n")
                    break


def experiment_stop(experimentid, remove, removefiles, **kwargs):
    ignore_checkpointing = kwargs.get('ignore_checkpointing', False)
    cmd = 'adapt-experiment-checkpoint -e %d' % (experimentid)
    rc = subprocess.call(cmd, shell=True)
    if rc:
        sys.stderr.write("ERROR: calling %s" % (cmd))
        return rc
    
    ssh = AdaptConfig.get_ssh_ssh()
    
    if ignore_checkpointing is False:
        while True:
            
            rows = adaptlibs.GetDataAccessObject().branches_runningSelectBranchIDWhere(expid=experimentid)
            
            if len(rows) == 0:
                break
            else:
                print('Waiting for checkpoint to complete, rows are', rows)
                time.sleep(5)
    elif ignore_checkpointing is True:
        rows = adaptlibs.GetDataAccessObject().branches_runningSelectBranchIDWhere(expid=experimentid)
        if len(rows) != 0:
            print('Attempting to delete experiment with branches marked as running. This may cause a mess.')
    
    rows = adaptlibs.GetDataAccessObject().experimentsSelectAllWhere(expid=experimentid)
    if len(rows) == 0:
        return
    
    relevant_branches = []
    rows = adaptlibs.GetDataAccessObject().branchesSelectAllWhere(expid=experimentid)
    for row in rows:
        relevant_branches.append(row[0])
    relevant_jobs = []
    job_directories_to_remove = []
    relevant_branches_set = ', '.join([repr(int(x)) for x in relevant_branches])
    if len(relevant_branches_set) > 0:
        rows = adaptlibs.GetDataAccessObject().jobsSelectAllWhereInBranchID(relevant_branches_set=relevant_branches_set)
        for row in rows:
            j_id, j_state, j_hostname, j_rundir, j_pid = row
            relevant_jobs.append(j_id)
            if j_state == 'R' and j_pid:
                kill_remote_pid_group(j_hostname, int(j_pid), True)
            if j_hostname and j_rundir:
                job_directories_to_remove.append((j_hostname, j_rundir))
    relevant_jobs_set = ', '.join([repr(int(x)) for x in relevant_jobs])
    print(experimentid)
    print(relevant_branches)
    print(relevant_jobs)
    
    if not remove:
        return
    
    localstoragedir = None
    if removefiles:
        rows = adaptlibs.GetDataAccessObject().experimentsSelectLocalStorageDirWhere(expid=experimentid)
        print('rows', rows)
        if rows:
            localstoragedir = rows[0][0]
    
    try:
        if relevant_branches_set:
            adaptlibs.GetDataAccessObject().branch_input_blobsDeleteAllWhereIn(relevant_branches_set=relevant_branches_set)
            adaptlibs.GetDataAccessObject().branch_input_refsDeleteAllWhereIn(relevant_branches_set=relevant_branches_set)
    except:
        print('An error occurred in adaptlibs.experiment_stop deleting branch_input_blobs or branch_input_refs.')
    
    try:
        if relevant_jobs_set:
            adaptlibs.GetDataAccessObject().job_output_blobsDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
            adaptlibs.GetDataAccessObject().job_output_infoDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
            adaptlibs.GetDataAccessObject().job_checkpoints_takenDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
            adaptlibs.GetDataAccessObject().job_output_lsDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
            adaptlibs.GetDataAccessObject().jobsDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
    except:
        print('An error occurred in adaptlibs.experiment_stop deleting jobs.')
    
    try:
        if relevant_branches_set:
            adaptlibs.GetDataAccessObject().branchesDeleteAllWhereIn(relevant_branches_set=relevant_branches_set)
    except:
        print('An error occurred in adaptlibs.experiment_stop deleting branches. Trying another (slower) method.')
        branches_remain = relevant_branches
        while len(branches_remain) > 0:
            for branchid in sorted(branches_remain):
                try:
                    adaptlibs.GetDataAccessObject().branchesDeleteAllWhere(branchid=branchid)
                except:
                    pass
            branches_remain = []
            rows_remain = adaptlibs.GetDataAccessObject().branchesSelectAllWhere(expid=experimentid)
            for row in rows_remain:
                branches_remain.append(row[0])
    
    try:
        adaptlibs.GetDataAccessObject().experimentsDeleteAllWhere(expid=experimentid)
        adaptlibs.GetDataAccessObject().experiment_environment_variablesDeleteAllWhere(expid=experimentid)
    except:
        print('An error occurred in adaptlibs.experiment_stop deleting the experiment.')
    
    host_directories = {}
    if removefiles:
        for host_directory in job_directories_to_remove:
            host, directory = host_directory
            if host not in host_directories:
                host_directories[host] = []
            host_directories[host].append(directory)
    
    #print('The following directories will be permanently deleted:\n')
    #print(host_directories)
    #resolved = False
    #while resolved is False:
    #    if int(sys.version[0]) == 2:
    #        really_delete = str(raw_input('Are you sure you want to permanently delete these directories? y or n.'))
    #    else:
    #        really_delete = str(input('Are you sure you want to permanently delete these directories? y or n.'))
    #    if really_delete == 'y':
    #        resolved = True
    #    elif really_delete == 'n':
    #        host_directories = {}
    #        resolved = True
    #    else:
    #        print('''Not a valid response. Please use "y" or "n".''')
    
    keys = sorted(host_directories.keys())
    for host in keys:
        print("cleanup on", host)
        cmd = '%s %s rm -fr %s' % (ssh, host, ' '.join([shell_escape(x) for x in host_directories[host]]))
        print('\-->  ', cmd)
        rc = subprocess.call(cmd, shell=True)
        if rc:
            for d in host_directories[host]:
                adaptlibs.GetDataAccessObject().pending_directories_to_removeInsertAllWhere(hostname=host, path=d)
    
    if localstoragedir:
        print('removing', localstoragedir)
        try:
            shutil.rmtree(localstoragedir)
        except OSError:
            pass


def branch_remove(bid_input, remove=False):
    # This function takes a branch and deletes it, its jobs, and all relevant entries from the database. It does not delete files. Child branches are not affected, and commands will fail if children exist.
    
    if remove is False:
        print('Either no argument was passed, or False was passed. Not deleting any data.')
    
    relevant_branches = []
    rows = adaptlibs.GetDataAccessObject().branchesSelectIDWhereBranchID(branchid=bid_input)
    if len(rows) == 0:
        return
    
    for row in rows:
        relevant_branches.append(row[0])
    relevant_jobs = []
    relevant_branches_set = ', '.join([repr(int(x)) for x in relevant_branches])
    rows = adaptlibs.GetDataAccessObject().jobsSelectAllWhereInBranchID(relevant_branches_set=relevant_branches_set)
    
    for row in rows:
        j_id, j_state, j_hostname, j_rundir, j_pid = row
        relevant_jobs.append(j_id)
        if j_state == 'R' and j_pid:
            kill_remote_pid_group(j_hostname, int(j_pid), True)
    relevant_jobs_set = ', '.join([repr(int(x)) for x in relevant_jobs])
    print(relevant_branches)
    print(relevant_jobs)
    
    if not remove:
        return
    
    try:
        if relevant_branches_set:
            try:
                print('Attempting to delete branch_input_blobs entry.')
                adaptlibs.GetDataAccessObject().branch_input_blobsDeleteAllWhereIn(relevant_branches_set=relevant_branches_set)
            except:
                print('Error running branch_input_blobsDeleteAllWhereIn. Branches not deleted.')
            try:
                print('Attempting to delete branch_input_refs entry.')
                adaptlibs.GetDataAccessObject().branch_input_refsDeleteAllWhereIn(relevant_branches_set=relevant_branches_set)
            except:
                print('Error running branch_input_refsDeleteAllWhereIn. Branches not deleted.')
            if relevant_jobs_set:
                try:
                    print('Attempting to delete job_output_blobs entry.')
                    adaptlibs.GetDataAccessObject().job_output_blobsDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
                except:
                    print('Error running job_output_blobsDeleteAllWhereIn. Branches not deleted.')
                try:
                    print('Attempting to delete job_output_info entry.')
                    adaptlibs.GetDataAccessObject().job_output_infoDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
                except:
                    print('Error running job_output_infoDeleteAllWhereIn. Branches not deleted.')
                try:
                    print('Attempting to delete job_checkpoints_taken entry.')
                    adaptlibs.GetDataAccessObject().job_checkpoints_takenDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
                except:
                    print('Error running job_checkpoints_takenDeleteAllWhereIn. Branches not deleted.')
                try:
                    print('Attempting to delete job_output_ls entry.')
                    adaptlibs.GetDataAccessObject().job_output_lsDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
                except:
                    print('Error running job_output_lsDeleteAllWhereIn. Branches not deleted.')
                try:
                    print('Attempting to delete jobs entry.')
                    adaptlibs.GetDataAccessObject().jobsDeleteAllWhereIn(relevant_jobs_set=relevant_jobs_set)
                except:
                    print('Error running jobsDeleteAllWhereIn. Branches not deleted.')
            try:
                print('Attempting to delete branches entry.')
                adaptlibs.GetDataAccessObject().branchesDeleteAllWhereIn(relevant_branches_set=relevant_branches_set)
            except:
                print('Error running branchesDeleteAllWhereIn. Branches likely not deleted.')
    except:
        print('An error occurred in adaptlibs.branch_remove.')


def experiment_copy(exp, exp_reduced_desc):
    # In this function, a copy of an experiment is created in the database.
    
    # Get the base experiment.
    base_experiment = adaptlibs.GetDataAccessObject().experimentsSelectAllWhere(expid=exp)
    
    # Create a new entry in experiments.
    adaptlibs.GetDataAccessObject().experimentsInsertAllAll(name=base_experiment[0][1], description=('%s reduced copy of %s' % (exp_reduced_desc, base_experiment[0][2])), probability_threshold=0.0, checkpointscript='null', localstoragedir='null', y_or_n='y')
    
    # Get all experiments again. Removal of a reduced copy depends on localstoragedir being null.
    all_experiments_2 = adaptlibs.GetDataAccessObject().experimentsSelectAllAll()
    
    # Get the experiment number of the copy to create and then reduce.
    reduced_experiment = all_experiments_2[-1][0]
    
    # Get branches associated with the base experiment.
    base_branches = adaptlibs.GetDataAccessObject().branchesSelectAllWhere(expid=exp)
    
    # Add the first branch, which has a parent of null.
    adaptlibs.GetDataAccessObject().branchesInsertAllRoot(expid=reduced_experiment, description='root', probability=base_branches[0][6], probability_threshold_multiplier=base_branches[0][7], terminate_early=base_branches[0][8], priority=base_branches[0][9], branchexecutable=base_branches[0][10])
    
    # Get that new branch's number.
    new_experiment_root_branch = adaptlibs.GetDataAccessObject().branchesSelectAllWhere(expid=reduced_experiment)
    new_experiment_root_branch = new_experiment_root_branch[0][0]
    
    # Make a list of parent branches, which will change from the base experiment with every new branch.
    new_parentid = [0] * len(base_branches)
    new_parentid_updated = [False] * len(base_branches)
    new_parentid_toadd = [False] * len(base_branches)
    new_branchid = [0] * len(base_branches)
    # We don't need to mess with the root branch again.
    new_parentid_updated[0] = True
    new_parentid_toadd[0] = False
    new_branchid[0] = new_experiment_root_branch
    # Populate the new list of parents with old values.
    for x in range(1, len(base_branches)):
        new_parentid[x] = int(base_branches[x][2])
    # Update the new parent branches array with the new root branch number.
    for x in range(1, len(base_branches)):
        if (int(base_branches[x][2]) == int(base_branches[0][0])):
            new_parentid[x] = int(new_experiment_root_branch)
            new_parentid_updated[x] = True
            new_parentid_toadd[x] = True
    
    # Loop over the rest of the branches. Continues while any branches are marked to add. This will naturally work from parent to child until the end.
    while any(new_parentid_toadd) is True:
        
        # Identify the index of the branch we will deal with.
        next_branch = next((i for i, v in enumerate(new_parentid_toadd) if v))
        
        # Add the new branch to branches.
        adaptlibs.GetDataAccessObject().branchesInsertAllChild(expid=reduced_experiment, parentid=new_parentid[next_branch], description=str(base_branches[next_branch][3]), probability=base_branches[next_branch][6], probability_threshold_multiplier=base_branches[next_branch][7], terminate_early=base_branches[next_branch][8], priority=base_branches[next_branch][9], branchexecutable=str(base_branches[next_branch][10]))
        
        # Get the new branch's number.
        new_experiment_branch = adaptlibs.GetDataAccessObject().branchesSelectIDWhereExpIDAndParentIDandDescription(expid=reduced_experiment, parentid=new_parentid[next_branch], description=base_branches[next_branch][3])
        new_experiment_branch = new_experiment_branch[0]
        new_branchid[next_branch] = new_experiment_branch
        
        # Update the new parent branches array with the new parent branch number. May not affect anything.
        for x in range(1, len(base_branches)):
            if (int(base_branches[x][2]) == int(base_branches[next_branch][0])):
                if type(new_experiment_branch) == tuple:
                    new_experiment_branch = new_experiment_branch[0]
                new_parentid[x] = new_experiment_branch
                new_parentid_updated[x] = True
                new_parentid_toadd[x] = True
        
        # Mark this branch as finished.
        new_parentid_toadd[next_branch] = False
    
    # Get jobs associated with the base experiment.
    base_jobs = adaptlibs.GetDataAccessObject().jobsSelectAllWhereExpIDCompleted(expid=exp)
    
    # Copy jobs from the base experiment, changing branchid.
    for x in range(0, len(base_jobs)):
        if type(new_branchid[x]) == int:
            use_this_branchid = new_branchid[x]
        else:
            use_this_branchid = new_branchid[x][0]
        adaptlibs.GetDataAccessObject().jobsInsertAllReduced(branchid=use_this_branchid, state=base_jobs[x][2], hostname=base_jobs[x][3], rundir=base_jobs[x][4])
    
    # Get jobs associated with the reduced experiment.
    reduced_jobs_ids = adaptlibs.GetDataAccessObject().jobsSelectAllWhereExpIDCompleted(expid=reduced_experiment)
    
    # Get job output information associated with the base experiment.
    base_job_output_info = adaptlibs.GetDataAccessObject().job_output_infoSelectInfoWhereFinishedByExpID(expid=base_experiment[0][0])
    
    # Map base jobid to reduced jobid for copying job_output_info.
    base_jobids = [0] * len(base_jobs)
    reduced_jobids = [0] * len(reduced_jobs_ids)
    for x in range(0, len(base_jobs)):
        base_jobids[x] = base_jobs[x][0]
        reduced_jobids[x] = reduced_jobs_ids[x][0]
    
    # Populate the base jobids.
    reduced_job_info_jobid = [0] * len(base_job_output_info)
    for x in range(0, len(base_job_output_info)):
        reduced_job_info_jobid[x] = base_job_output_info[x][0]
    
    # Loop through all lines of reduced_job_info_jobid, replacing jobid.
    for x in range(0, len(reduced_job_info_jobid)):
        for y in range(0, len(reduced_jobids)):
            if reduced_job_info_jobid[x] == base_jobids[y]:
                reduced_job_info_jobid[x] = reduced_jobids[y]
    
    # Copy job_output_info from the base experiment, changing jobid.
    for x in range(0, len(base_job_output_info)):
        adaptlibs.GetDataAccessObject().job_output_infoInsertAllWhere(jobid=reduced_job_info_jobid[x], k=base_job_output_info[x][1], v=base_job_output_info[x][2])
    
    # Copy branch_input_blobs from the base experiment, changing branchid.
    relevant_branches_set = ', '.join([str(x[0]) for x in base_branches])
    base_branch_input_blobs = adaptlibs.GetDataAccessObject().branch_input_blobsSelectAllWhereInBranchID(relevant_branches_set=relevant_branches_set)
    # Propagate new branch numbers.
    new_branch_input_blobs = [] * len(base_branch_input_blobs)
    for x in range(0, len(base_branch_input_blobs)):
        new_branch_input_blobs.append(list(base_branch_input_blobs[x]))
    for x in range(0, len(new_branch_input_blobs)):
        oldbranch = new_branch_input_blobs[x][0]
        old_index = [idx for idx in range(len(base_branches)) if base_branches[idx][0] == oldbranch][0]
        newbranch = new_branchid[old_index][0]
        new_branch_input_blobs[x][0] = newbranch
    
    # Insert the new branch_input_blobs entries.
    if int(sys.version[0]) == 2:
        for line in new_branch_input_blobs:
            adaptlibs.GetDataAccessObject().branch_input_blobsInsert(branchid=line[0], filename=line[1], permissions=line[2], dat=line[3])
    else:
        for line in new_branch_input_blobs:
            adaptlibs.GetDataAccessObject().branch_input_blobsInsert(branchid=line[0], filename=line[1], permissions=line[2], dat=line[3].decode())
    # Return base experiment number, description to append to copy, copy experiment number, whole record of copy experiment
    new_experiment_row = all_experiments_2[-1]
    return (exp, exp_reduced_desc, reduced_experiment, new_experiment_row)


def get_branch_heritage(bid):
    out = []
    
    while True:
        out.insert(0, int(bid))
        val = adaptlibs.GetDataAccessObject().branchesSelectParentIDWhereBranchID(branchid=bid)
        if val:
            val = val[0][0]
        if not val:
            break
        bid = val
    return out


def get_exp_branch_heritage(expid, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    # This function was added because it is sometimes desired to get the heritage for every branch in an experiment. Calling the database for every generation of every sequence as in get_branch_heritage is very slow. This function makes one database call and does the math.
    
    heritage_pickle_filename = 'branch_heritage_%s.pick' % (str(expid))
    heritage_json_filename = 'branch_heritage_%s.json' % (str(expid))
    plot_cache_path = os.path.join(AdaptConfig.get_server_adaptemp(), 'plot_cache')
    
    # If we're ok with using old data and it exists, load it and be done.
    if (use_old_data is True) and (os.path.isfile(os.path.join(plot_cache_path, heritage_json_filename))):
        print('Per availability and input argument, get_exp_branch_heritage is using old heritage data.')
        if int(sys.version[0]) == 2:
            json_branch_heritage = json.load(open(os.path.join(plot_cache_path, heritage_json_filename), 'rb'))
        else:
            json_branch_heritage = json.load(open(os.path.join(plot_cache_path, heritage_json_filename), 'r'))
        expid = json_branch_heritage['expid']
        root_branch = json_branch_heritage['root_branch']
        end_branches = json_branch_heritage['end_branches']
        exp_branch_heritage = {int(x): json_branch_heritage['exp_branch_heritage'][x] for x in json_branch_heritage['exp_branch_heritage']}
        return (expid, root_branch, end_branches, exp_branch_heritage)
    elif (use_old_data is True) and (os.path.isfile(os.path.join(plot_cache_path, heritage_pickle_filename))):
        print('Per availability and input argument, get_exp_branch_heritage is using old heritage data.')
        if int(sys.version[0]) == 2:
            pickled_branch_heritage = pickle.load(open(os.path.join(plot_cache_path, heritage_pickle_filename), 'rb'))
        else:
            pickled_branch_heritage = pickle.load(open(os.path.join(plot_cache_path, heritage_pickle_filename), 'r'))
        expid = pickled_branch_heritage['expid']
        root_branch = pickled_branch_heritage['root_branch']
        end_branches = pickled_branch_heritage['end_branches']
        exp_branch_heritage = {int(x): pickled_branch_heritage['exp_branch_heritage'][x] for x in pickled_branch_heritage['exp_branch_heritage']}
        return (expid, root_branch, end_branches, exp_branch_heritage) 
    
    # Get branch relationships and identify root branch.
    ids_and_parentids = adaptlibs.GetDataAccessObject().branchesSelectIDParentIDWhereExpID(expid=expid)
    try:
        root_branch = [x[0] for x in ids_and_parentids if (x[1] is None)][0]
    except IndexError:
        print('There appear to be no branches under experimentid %s.' % str(expid))
    
    # Gather all unique parents, and all end branches (non-parents).
    parents = set([x[1] for x in ids_and_parentids if (x[1] is not None)])
    parent_dict = {x[1]: x[0] for x in ids_and_parentids if (x[0] is not None)}
    child_dict = {x[0]: x[1] for x in ids_and_parentids if (x[1] is not None)}
    allbranches = set([x[0] for x in ids_and_parentids if (x[0] is not None)])
    end_branches = list(allbranches - parents)
    
    # For all non-root branches, build a heritage until the root branch is reached.
    exp_branch_heritage = {}
    for branches in ids_and_parentids:
        exp_branch_heritage[branches[0]] = []
        # First in the heritage is the branch itself.
        exp_branch_heritage[branches[0]].append(branches[0])
        last = branches[0]
        while last != root_branch:
            # Find the parent of branches[0]
            par = child_dict[last]
            exp_branch_heritage[branches[0]].append(par)
            last = par
        exp_branch_heritage[branches[0]].reverse()
    
    # Check if the plot_cache directory exists.
    plot_cache_path = os.path.join(AdaptConfig.get_server_adaptemp(), 'plot_cache')
    if os.path.isdir(plot_cache_path):
        pass
    else:
        os.makedirs(plot_cache_path)
    
    # Assemble the heritage to the plot_cache directory.
    json_branch_heritage = {}
    json_branch_heritage['expid'] = expid
    json_branch_heritage['root_branch'] = root_branch
    json_branch_heritage['end_branches'] = end_branches
    json_branch_heritage['exp_branch_heritage'] = exp_branch_heritage
    
    # Dump a copy of the heritage to the plot_cache directory.
    if (sys.version[0]) == 2:
        json.dump(json_branch_heritage, open(os.path.join(plot_cache_path, heritage_json_filename), 'wb'))
    else:
        json.dump(json_branch_heritage, open(os.path.join(plot_cache_path, heritage_json_filename), 'w'))
    
    return (expid, root_branch, end_branches, exp_branch_heritage)


def get_plot_data(plot_gather_input):
    # Given a directory/file path, this function retrieves two separated columns of data.
    # Check that plot file exists and contains data.
    ssh = AdaptConfig.get_ssh_ssh()
    
    host = plot_gather_input[0]
    plotfile = plot_gather_input[1]
    
    cmd = """%s %s 'cat %s;'""" % (ssh, host, str(plotfile))
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    stdout, stderr = p.communicate()
    if len(stderr) > 0:
        print('An error occurred as follows: %s' % stderr)
    if len(stdout) == 0:
        return False
    
    print('Processing plot data for host:file %s:%s.' % (host, plotfile))
    if int(sys.version[0]) == 2:
        lines = stdout.replace('\r\n', '\n')
        lines = lines.split('\n')
        while (lines[-1] == '\r\n') or (lines[-1] == '\n') or (lines[-1] == ''):
            lines.pop(-1)
        separator = ''
        if ',' in lines[0]:
            separator = ','
        if ':' in lines[0]:
            separator = ':'
        if ';' in lines[0]:
            separator = ';'
    else:
        lines = stdout.replace(b'\r\n', b'\n')
        lines = lines.split(b'\n')
        while (lines[-1] == b'\r\n') or (lines[-1] == b'\n') or (lines[-1] == b''):
            lines.pop(-1)
        separator = b''
        if b',' in lines[0]:
            separator = b','
        if b':' in lines[0]:
            separator = b':'
        if b';' in lines[0]:
            separator = b';'
    plot_time = [0.0] * len(lines)
    plot_value = [0.0] * len(lines)
    
    for k in range(0, len(lines)):
        try:
            if len(separator) > 0:
                plot_time[k] = float((lines[k].split(separator))[0])
                plot_value[k] = float((lines[k].split(separator))[1])
            else:
                plot_time[k] = float((lines[k].split())[0])
                plot_value[k] = float((lines[k].split())[1])
        except ValueError:
            print('There does not appear to be valid plot data in host:file %s:%s. Skipping.' % (host, plotfile))
            return False
        combined_list = [list(a) for a in zip(plot_time, plot_value)]
    return combined_list


def get_branch_plot_data(branchid, plotfile):
    # Given a branchid and plotfile name, this function retrieves two separated columns of data.
    ssh = AdaptConfig.get_ssh_ssh()
    
    job = adaptlibs.GetDataAccessObject().jobsSelectAllWhereStateAndBranchID(branchid=branchid, state='C')
    if len(job) == 0:
        return False
    
    host = str(job[0][3])
    plotfile = os.path.join(str(job[0][4]), str(plotfile))
    combined_list = adaptlibs.get_plot_data([host, plotfile])
    return combined_list


def integrate_plot_data(plot_values):
    # Some plot data returns a rate which is may be useful to integrate.
    # Currently uses the trapezoidal rule.
    time_series = [x[0] for x in plot_values]
    valu_series = [x[1] for x in plot_values]
    import scipy
    cum_value = scipy.integrate.cumtrapz(valu_series, time_series)
    cum_value = list(cum_value)
    new_combined_list = [list(a) for a in zip(time_series, cum_value)]
    return new_combined_list


def get_heritage_plot_data(plot_gather_input):
    # Returns plot data for a desired plotfile for a desired branch up to the root branch.
    branchid = plot_gather_input[0]
    plotfile = plot_gather_input[1]
    heritage = adaptlibs.get_branch_heritage(branchid)
    raw_plot_values = []
    for branch in heritage:
        print('Checking branchid %s for plotfile %s.' % (str(branch), plotfile))
        if adaptlibs.get_branch_plot_data(branch, plotfile):
            raw_plot_values.extend(adaptlibs.get_branch_plot_data(branch, plotfile))
    # Remove duplicate data.
    plot_values_unsorted = [list(t) for t in set(tuple(element) for element in raw_plot_values)]
    # Sort by time.
    plot_values = sorted(plot_values_unsorted, key=lambda x: x[0])
    try:
        plot_values
    except:
        plot_values = False
    return plot_values


def check_jobdir_for_plotfile(check_package):
    # This searches a job directory for a given file name. Refactored for use in parallel.
    branchid = check_package[0]
    if int(sys.version[0]) == 2:
        plotfile = check_package[1]
        host = check_package[2]
        jobdir = check_package[3]
    else:
        plotfile = str.encode(check_package[1])
        host = str.encode(check_package[2])
        jobdir = str.encode(check_package[3])
    
    print('Checking jobdir %s on host %s for plotfile %s.' % (jobdir, host, plotfile))
    ssh = AdaptConfig.get_ssh_ssh()
    if int(sys.version[0]) == 2:
        cmd = """%s %s 'ls -la %s;'""" % (ssh, host, jobdir)
    else:
        cmd = """%s %s 'ls -la %s;'""" % (ssh, host.decode(), jobdir.decode())
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    stdout, stderr = p.communicate()
    if int(sys.version[0]) == 2:
        filelist_raw = stdout.split('\n')
        for text in filelist_raw:
            if plotfile in text:
                plot_file_meta = {}
                plot_file_meta['file'] = ' '.join(text.split()[8:])
                plot_file_meta['mdate'] = ' '.join(text.split()[5:8])
                plot_file_meta['size'] = text.split()[4]
    else:
        filelist_raw = stdout.split(b'\n')
        for text in filelist_raw:
            if plotfile in text:
                plot_file_meta = {}
                plot_file_meta['file'] = (b' '.join(text.split()[8:])).decode()
                plot_file_meta['mdate'] = (b' '.join(text.split()[5:8])).decode()
                plot_file_meta['size'] = (text.split()[4]).decode()
    try:
        plot_file_meta
    except:
        plot_file_meta = False
    return (branchid, plot_file_meta)


def get_exp_heritage_plot_data(expid, plotfile, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    use_processors = kwargs.get('use_processors', 1)
    # Returns plot data for a desired plotfile for all end branches of an experiment up to the root branch.
    (expid, root_branch, end_states, exp_branch_heritage) = adaptlibs.get_exp_branch_heritage(expid, use_old_data=use_old_data)
    branches_to_jobs = adaptlibs.GetDataAccessObject().jobsSelectAllWhereExpIDCompleted(expid=expid)
    
    checker_input = []
    for job in branches_to_jobs:
        checker_input.append([])
        checker_input[-1].append(job[1])
        checker_input[-1].append(plotfile)
        checker_input[-1].append(job[3])
        checker_input[-1].append(job[4])
    
    # This will cache plot file outputs demanded by adaptlibs.get_exp_heritage_plot_data for ADAPT experiments, saving time when calculating multiple DYIs or other analyses. It should be added to the experiment deletion process.
    # The branch directories of the experiment are searched for the given plot file. The directory names, modified dates, and sizes of plot files are examined. If there is no previous set of such information or if it has changed, the current set of metadata as well as the actual plot data are saved for future use.
    
    plot_file_meta = {}
    
    # Check if the plot_cache directory exists.
    plot_cache_path = os.path.join(AdaptConfig.get_server_adaptemp(), 'plot_cache')
    if os.path.isdir(plot_cache_path):
        pass
    else:
        os.makedirs(plot_cache_path)
    
    meta_pickle_filename = '%s_%s_meta.pick' % (plotfile, str(expid))
    data_pickle_filename = '%s_%s_vals.pick' % (plotfile, str(expid))
    meta_json_filename = '%s_%s_meta.json' % (plotfile, str(expid))
    data_json_filename = '%s_%s_vals.json' % (plotfile, str(expid))
    
    # If use_old_data is True, look for old meta data and save some time. Otherwise, proceed as if it is the first time.
    if ((use_old_data is True) and (os.path.isfile(os.path.join(plot_cache_path, meta_json_filename)))):
        print('Per availability and input argument, get_exp_heritage_plot_data is using old meta data.')
        if int(sys.version[0]) == 2:
            plot_file_meta = json.load(open(os.path.join(plot_cache_path, meta_json_filename), 'rb'))
        else:
            plot_file_meta = json.load(open(os.path.join(plot_cache_path, meta_json_filename), 'r'))
        changed = False
    elif ((use_old_data is True) and (os.path.isfile(os.path.join(plot_cache_path, meta_pickle_filename)))):
        print('Per availability and input argument, get_exp_heritage_plot_data is using old meta data.')
        if int(sys.version[0]) == 2:
            plot_file_meta = pickle.load(open(os.path.join(plot_cache_path, meta_pickle_filename), 'rb'))
        else:
            plot_file_meta = pickle.load(open(os.path.join(plot_cache_path, meta_pickle_filename), 'r'))
        changed = False
    else:
        print('Checking for plot files under directories associated with expid %s. Total %s branches.' % (str(expid), len(branches_to_jobs)))
        if use_processors < 2:
            for x, check_package in enumerate(checker_input):
                print('Checking branch %s of %s.' % (str(x + 1), len(branches_to_jobs)))
                branchid, meta = adaptlibs.check_jobdir_for_plotfile(check_package)
                plot_file_meta[branchid] = meta
        else:
            from multiprocessing import Pool as ProcessPool
            mypool = ProcessPool(use_processors)
            results = mypool.map(adaptlibs.check_jobdir_for_plotfile, checker_input)
            for line in results:
                plot_file_meta[line[0]] = line[1]
            del results
            mypool.close()
            mypool.join()
        
        # If the metadata file already exists in the cache directory, check if it is the same as the data just gathered. If they are the same, we can use the associated data file. If not, continue in the function to gather plot data and eventually save it there for next time.
        if os.path.isfile(os.path.join(plot_cache_path, data_json_filename)):
            try:
                if int(sys.version[0]) == 2:
                    old_plot_file_meta = json.load(open(os.path.join(plot_cache_path, meta_json_filename), 'rb'))
                else:
                    old_plot_file_meta = json.load(open(os.path.join(plot_cache_path, meta_json_filename), 'r'))
                if old_plot_file_meta == plot_file_meta:
                    changed = False
                else:
                    # Old metadata json exists but is different.
                    changed = True
                    if int(sys.version[0]) == 2:
                        json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'wb'))
                    else:
                        json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'w'))
            except:
                # No old metadata json exists or it cannot be read.
                changed = True
                if int(sys.version[0]) == 2:
                    json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'wb'))
                else:
                    json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'w'))
                
        elif os.path.isfile(os.path.join(plot_cache_path, data_pickle_filename)):
            try:
                old_plot_file_meta = pickle.load(open(os.path.join(plot_cache_path, meta_pickle_filename), 'rb'))
                if old_plot_file_meta == plot_file_meta:
                    changed = False
                else:
                    # Old metadata pickle exists but is different.
                    changed = True
                    if int(sys.version[0]) == 2:
                        json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'wb'))
                    else:
                        json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'w'))
            except:
                # No old metadata pickle exists or it cannot be read.
                changed = True
                if int(sys.version[0]) == 2:
                    json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'wb'))
                else:
                    json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'w'))
        else:
            # No old metadata exists or it cannot be read.
            changed = True
            if int(sys.version[0]) == 2:
                json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'wb'))
            else:
                json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'w'))
    
    # If there is no change, load the old data and save some time. Otherwise, proceed as if it is the first time.
    if ((use_old_data is True) or (changed is False)) and (os.path.isfile(os.path.join(plot_cache_path, data_json_filename))):
        print('Per availability and input argument (or by virtue of the data remaining unchanged), get_exp_heritage_plot_data is using old plot data.')
        if int(sys.version[0]) == 2:
            plot_values = json.load(open(os.path.join(plot_cache_path, data_json_filename), 'rb'))
            plot_values = {int(x): plot_values[x] for x in plot_values}
            return plot_values
        else:
            plot_values = json.load(open(os.path.join(plot_cache_path, data_json_filename), 'r'))
            plot_values = {int(x): plot_values[x] for x in plot_values}
            return plot_values
    elif ((use_old_data is True) or (changed is False)) and (os.path.isfile(os.path.join(plot_cache_path, data_pickle_filename))):
        print('Per availability and input argument (or by virtue of the data remaining unchanged), get_exp_heritage_plot_data is using old plot data.')
        if int(sys.version[0]) == 2:
            plot_values = pickle.load(open(os.path.join(plot_cache_path, data_pickle_filename), 'rb'))
            return plot_values
        else:
            plot_values = pickle.load(open(os.path.join(plot_cache_path, data_pickle_filename), 'r'))
            return plot_values
    else:
        print('Per availability and/or input argument, get_exp_heritage_plot_data is pulling new plot data. This may take some time.')
        
    # Assemble plot file gathering into a single list for use in parallel.
    heritage_plot_input = []
    for end in end_states:
        heritage_plot_input.append([])
        heritage_plot_input[-1].append(end)
        heritage_plot_input[-1].append(plotfile)
    
    plot_values = {}
    print('Total end states: %s' % len(end_states))
    plot_gather_input = []
    plot_gather_associate = []
    for branch in checker_input:
        if branch[0] in plot_file_meta:
            plot_gather_input.append([])
            plot_gather_input[-1].append(branch[2])
            plot_gather_input[-1].append(os.path.join(branch[3], branch[1]))
            plot_gather_associate.append(branch[0])
    
    if use_processors < 2:
        results = []
        for x, branch in enumerate(plot_gather_input):
            print('Gathering data for branch %s of %s.' % (str(x + 1), len(plot_gather_input)))
            results.append(adaptlibs.get_plot_data(branch))
    else:
        # Gather plot data without regard for branch relationships.
        from multiprocessing import Pool as ProcessPool
        mypool = ProcessPool(use_processors)
        results = mypool.map(adaptlibs.get_plot_data, plot_gather_input)
        mypool.close()
        mypool.join()
    
    # Structure the plot data by branch.
    unrelated_plot_data = {}
    for x, line in enumerate(results):
        branchnum = plot_gather_associate[x]
        unrelated_plot_data[branchnum] = line
    
    del results
    
    # Structure the plot data by end state.
    plot_values = {}
    
    for end in end_states:
        plot_values[end] = []
        for heritage in exp_branch_heritage[end]:
            if (heritage in unrelated_plot_data) and (unrelated_plot_data[heritage]):
                plot_values[end].extend(unrelated_plot_data[heritage])
        # Remove duplicate data.
        plot_values[end] = [list(t) for t in set(tuple(element) for element in plot_values[end])]
        # Sort by time.
        plot_values[end] = sorted(plot_values[end], key=lambda x: x[0])
        try:
            dummyvariable = plot_values[end]
        except:
            plot_values[end] = False
        if len(plot_values[end]) == 0:
            plot_values[end] = False
    
    # Dump to a json for future use.
    if int(sys.version[0]) == 2:
        json.dump(plot_values, open(os.path.join(plot_cache_path, data_json_filename), 'wb'))
    else:
        json.dump(plot_values, open(os.path.join(plot_cache_path, data_json_filename), 'w'))
    return plot_values


def get_exp_unsorted_plot_data(expid, plotfile, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    use_processors = kwargs.get('use_processors', 1)
    # Returns plot data for a desired plotfile for all branches of an experiment. Let someone else sort em out.
    branches_to_jobs = adaptlibs.GetDataAccessObject().jobsSelectAllWhereExpIDCompleted(expid=expid)
    checker_input = []
    for job in branches_to_jobs:
        checker_input.append([])
        checker_input[-1].append(job[1])
        checker_input[-1].append(plotfile)
        checker_input[-1].append(job[3])
        checker_input[-1].append(job[4])
    
    # This will cache plot file outputs demanded by adaptlibs.get_exp_heritage_plot_data for ADAPT experiments, saving time when calculating multiple DYIs or other analyses. It should be added to the experiment deletion process.
    # The branch directories of the experiment are searched for the given plot file. The directory names, modified dates, and sizes of plot files are examined. If there is no previous set of such information or if it has changed, the current set of metadata as well as the actual plot data are saved for future use.
    
    plot_file_meta = {}
    
    # Check if the plot_cache directory exists.
    plot_cache_path = os.path.join(AdaptConfig.get_server_adaptemp(), 'plot_cache')
    if os.path.isdir(plot_cache_path):
        pass
    else:
        os.makedirs(plot_cache_path)
    
    meta_json_filename = '%s_%s_unsorted_meta.json' % (plotfile, str(expid))
    data_json_filename = '%s_%s_unsorted_vals.json' % (plotfile, str(expid))
    
    # If use_old_data is True, look for old meta data and save some time. Otherwise, proceed as if it is the first time.
    if ((use_old_data is True) and (os.path.isfile(os.path.join(plot_cache_path, meta_json_filename)))):
        print('Per availability and input argument, get_exp_heritage_plot_data is using old meta data.')
        if int(sys.version[0]) == 2:
            plot_file_meta = json.load(open(os.path.join(plot_cache_path, meta_json_filename), 'rb'))
        else:
            plot_file_meta = json.load(open(os.path.join(plot_cache_path, meta_json_filename), 'r'))
        changed = False
    else:
        print('Checking for plot files under directories associated with expid %s. Total %s branches.' % (str(expid), len(branches_to_jobs)))
        if use_processors < 2:
            for x, check_package in enumerate(checker_input):
                print('Checking branch %s of %s.' % (str(x + 1), len(branches_to_jobs)))
                branchid, meta = adaptlibs.check_jobdir_for_plotfile(check_package)
                plot_file_meta[branchid] = meta
        else:
            from multiprocessing import Pool as ProcessPool
            mypool = ProcessPool(use_processors)
            results = mypool.map(adaptlibs.check_jobdir_for_plotfile, checker_input)
            for line in results:
                plot_file_meta[line[0]] = line[1]
            del results
            mypool.close()
            mypool.join()
        
        # If the metadata file already exists in the cache directory, check if it is the same as the data just gathered. If they are the same, we can use the associated data file. If not, continue in the function to gather plot data and eventually save it there for next time.
        if os.path.isfile(os.path.join(plot_cache_path, data_json_filename)):
            try:
                if int(sys.version[0]) == 2:
                    old_plot_file_meta = json.load(open(os.path.join(plot_cache_path, meta_json_filename), 'rb'))
                else:
                    old_plot_file_meta = json.load(open(os.path.join(plot_cache_path, meta_json_filename), 'r'))
                if old_plot_file_meta == plot_file_meta:
                    changed = False
                else:
                    # Old metadata json exists but is different.
                    changed = True
                    if int(sys.version[0]) == 2:
                        json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'wb'))
                    else:
                        json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'w'))
            except:
                # No old metadata json exists or it cannot be read.
                changed = True
                if int(sys.version[0]) == 2:
                    json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'wb'))
                else:
                    json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'w'))
                
        else:
            # No old metadata exists or it cannot be read.
            changed = True
            if int(sys.version[0]) == 2:
                json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'wb'))
            else:
                json.dump(plot_file_meta, open(os.path.join(plot_cache_path, meta_json_filename), 'w'))
    
    # If there is no change, load the old data and save some time. Otherwise, proceed as if it is the first time.
    if ((use_old_data is True) or (changed is False)) and (os.path.isfile(os.path.join(plot_cache_path, data_json_filename))):
        print('Per availability and input argument (or by virtue of the data remaining unchanged), get_exp_heritage_plot_data is using old plot data.')
        if int(sys.version[0]) == 2:
            plot_values = json.load(open(os.path.join(plot_cache_path, data_json_filename), 'rb'))
            plot_values = {int(x): plot_values[x] for x in plot_values}
            return plot_values
        else:
            plot_values = json.load(open(os.path.join(plot_cache_path, data_json_filename), 'r'))
            plot_values = {int(x): plot_values[x] for x in plot_values}
            return plot_values
    else:
        print('Per availability and/or input argument, get_exp_heritage_plot_data is pulling new plot data. This may take some time.')
    
    plot_values = {}
    print('Total branches: %s' % len(plot_file_meta))
    plot_gather_input = []
    plot_gather_associate = []
    for branch in checker_input:
        if branch[0] in plot_file_meta:
            plot_gather_input.append([])
            plot_gather_input[-1].append(branch[2])
            plot_gather_input[-1].append(os.path.join(branch[3], branch[1]))
            plot_gather_associate.append(branch[0])
    
    if use_processors < 2:
        results = []
        for x, branch in enumerate(plot_gather_input):
            print('Gathering data for branch %s of %s.' % (str(x + 1), len(plot_gather_input)))
            results.append(adaptlibs.get_plot_data(branch))
    else:
        # Gather plot data without regard for branch relationships.
        from multiprocessing import Pool as ProcessPool
        mypool = ProcessPool(use_processors)
        results = mypool.map(adaptlibs.get_plot_data, plot_gather_input)
        mypool.close()
        mypool.join()
    
    # Structure the plot data by branch.
    plot_values = {}
    for x, line in enumerate(results):
        branchnum = plot_gather_associate[x]
        plot_values[branchnum] = line
    
    del results
    
    # Dump to a json for future use.
    if int(sys.version[0]) == 2:
        json.dump(plot_values, open(os.path.join(plot_cache_path, data_json_filename), 'wb'))
    else:
        json.dump(plot_values, open(os.path.join(plot_cache_path, data_json_filename), 'w'))
    return plot_values


# Dump the combined data from get_exp_heritage_plot_data into a pickle file in a chosen directory. Will not raise error on missing data or bad plot file name.
def pickle_exp_heritage_plot_data(expid, plotfile, pickle_location, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    use_processors = kwargs.get('use_processors', 1)
    heritage_vals = adaptlibs.get_exp_heritage_plot_data(expid, plotfile, use_old_data=use_old_data, use_processors=use_processors)
    filename = plotfile + '_' + str(expid) + '.pick'
    fullpath = os.path.join(pickle_location, filename)
    try:
        pickle.dump(heritage_vals, open(fullpath, 'wb'))
    except IOError:
        print('Unable to create pickled plot file. Check permissions in directory %s.' % pickle_location)


# Dump the combined data from get_exp_heritage_plot_data into a json file in a chosen directory. Will not raise error on missing data or bad plot file name.
def json_exp_heritage_plot_data(expid, plotfile, json_location, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    use_processors = kwargs.get('use_processors', 1)
    heritage_vals = adaptlibs.get_exp_heritage_plot_data(expid, plotfile, use_old_data=use_old_data, use_processors=use_processors)
    filename = plotfile + '_' + str(expid) + '.json'
    fullpath = os.path.join(json_location, filename)
    try:
        json.dump(heritage_vals, open(fullpath, 'wb'))
    except IOError:
        print('Unable to create json plot file. Check permissions in directory %s.' % json_location)


# Identify which end branches have finished and which have not.
def identify_finished_sequences(experimentid, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    
    (experimentid, root_branch, end_states, exp_branch_heritage) = adaptlibs.get_exp_branch_heritage(experimentid, use_old_data=use_old_data)
    all_completed = adaptlibs.GetDataAccessObject().branches_completedSelectAllWhere(expid=experimentid)
    all_completed = [x[1] for x in all_completed]
    
    finished_sequences = []
    unfinished_sequences = []
    for end in end_states:
        if end in all_completed:
            finished_sequences.append(end)
        else:
            unfinished_sequences.append(end)
    
    return (end_states, finished_sequences, unfinished_sequences)


def get_safe_branches_to_move(expid, move_commit, **kwargs):
    sb_remote_server = kwargs.get('remote_server', '')
    sb_remote_user = kwargs.get('remote_user', '')
    sb_remote_base_dir = kwargs.get('remote_base_dir', '')
    sb_branch_limit = kwargs.get('branch_limit', 1.E20)
    # This will identify branch directories that are safe to move or delete. In this case "safe" means the branch has finished and all of its children have finished. This is only useful in cases with very little storage space.
    
    # Retrieve all finished branches.
    all_completed = adaptlibs.GetDataAccessObject().branches_completedSelectAllWhere(expid=expid)
    all_completed_list = [x[1] for x in all_completed]
    all_completed_dict = {}
    for branch in all_completed:
        all_completed_dict[branch[1]] = branch[4]

    # Find the children of each completed branch.
    all_completed_children = {}
    for branch in all_completed:
        result = adaptlibs.GetDataAccessObject().branchesSelectIDWhereParentID(parentid=branch[1])
        all_completed_children[branch[1]] = result
    
    # Cycle through all completed branches and see if they have incomplete children. If they do, they are not considered safe to move.
    safe_to_move = {}
    for branch in all_completed_children:
        # Mark all as safe to begin.
        safe_to_move[branch] = True
    for branch in all_completed_children:
        # If there are no children, the branch is assumed safe.
        if len(all_completed_children[branch]) > 0:
            # If there are children, we check if they are finished.
            for child in all_completed_children[branch]:
                if not(child[0] in all_completed_list):
                    # If any children are not completed, the branch is not safe to move.
                    safe_to_move[branch] = False
                    break
    err = False
    if move_commit is True:
        if not sb_remote_server:
            print("If moving branches, a parameter named remote_server must be passed in the form: remote_server='remotehost'.")
            err = True
        if not sb_remote_user:
            print("If moving branches, a parameter named remote_user must be passed in the form: remote_user='remoteuser'.")
            err = True
        if not sb_remote_base_dir:
            print("If moving branches, a parameter named remote_base_dir must be passed in the form: remote_base_dir='/home/user-name/adapt-copy-dir/'.")
            err = True
        sb_branch_limit = int(sb_branch_limit)
    if err is True:
        Exception('There is an input error in adaptlibs.get_safe_branches_to_move.')
    
    # This snippet of code will move the branches that are safe and delete the original.
    if move_commit is True:
        deletions = 0
        for branch in safe_to_move:
            if deletions <= sb_branch_limit:
                if safe_to_move[branch] is True:
                    try:
                        # Assemble a command to move the directory. Assumes there is a password-less SSH/SCP connection.
                        orig_dir = all_completed_dict[branch]
                        move_command = ("scp -r " + orig_dir + " " + sb_remote_user + "@" + sb_remote_server + ":" + sb_remote_base_dir)
                        process = subprocess.Popen(move_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        stdout, stderr = process.communicate()
                        # Delete the original.
                        shutil.rmtree(orig_dir)
                        deletions += 1
                    except:
                        pass
    return safe_to_move


def find_var_relevant_branches_and_end_states(expid, bc_chosen, var_chosen, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    # This finds end state branches that have had the chosen branching condition occur AND have had the chosen variable changed from its initial value.
    
    # Get all completed jobs and branches from the experiment.
    finished_job_info = adaptlibs.GetDataAccessObject().job_output_infoSelectInfoWhereFinishedByExpID(expid=expid)
    relevant_branching_jobs = [row for row in finished_job_info if str(row[1]) == 'code' and str(row[2]) == str(bc_chosen)]
    if len(relevant_branching_jobs) == 0:
        raise ValueError('It appears that branching condition %s never occurs in experiment %s. find_var_relevant_branches_and_end_states has failed.' % (bc_chosen, expid))
    
    # Get a mapping of jobs and branches for the chosen experiment. Only includes finished branches and jobs.
    jobs_and_branches = adaptlibs.GetDataAccessObject().jobsSelectAllWhereExpIDCompleted(expid=expid)
    
    # Translate jobs to branches. This yields a list of branches where the chosen branching condition was used.
    relevant_branches = []
    for x in range(0, len(relevant_branching_jobs)):
        for y in range(0, len(jobs_and_branches)):
            if jobs_and_branches[y][0] == relevant_branching_jobs[x][0]:
                relevant_branches.append(jobs_and_branches[y][1])
    
    # Find the immediate children of those branches.
    relevant_branch_children = {}
    for x in range(0, len(relevant_branches)):
        relevant_branch_children[str(relevant_branches[x])] = adaptlibs.children_identify(relevant_branches[x])
    
    # Gather the editrule state of each child branch and see what value was given for var_chosen.
    relevant_child_branches_set = []
    for x in relevant_branches:
        for y in relevant_branch_children[str(x)]:
            if y not in relevant_child_branches_set:
                relevant_child_branches_set.append(y)
    
    relevant_child_branches_set = ', '.join([repr(int(x)) for x in relevant_child_branches_set])
    branch_var_inputs_all = adaptlibs.GetDataAccessObject().branch_input_blobsSelectAllWhereInBranchIDandState(relevant_branches_set=relevant_child_branches_set)
    
    child_branch_values = {}
    for x in relevant_branches:
        for y in relevant_branch_children[str(x)]:
            for z in branch_var_inputs_all:
                if str(z[0]) == str(y):
                    branch_var_inputs = z
                    # Return the file that contains ".state" in its name as the editrules pickle.
                    branch_var_inputs = [k for k in branch_var_inputs]
                    if len(branch_var_inputs) > 0:
                        branch_var_input_data = pickle.loads(z[3])
                        # This returns the value of the variable var_chosen for this branch.
                        branch_var_value = branch_var_input_data[1][var_chosen]
                        child_branch_values[str(y)] = branch_var_value
    
    # Find all end branches.
    # Reduce the list of all end branches to end branches relevant to the branching condition.
    (expid, root_branch, end_states, exp_branch_heritage) = adaptlibs.get_exp_branch_heritage(expid, use_old_data=use_old_data)
    
    relevant_end_states = []
    for x in end_states:
        # If a member of relevant_branches is in the heritage
        if not(set(relevant_branches).isdisjoint(exp_branch_heritage[x])):
            relevant_end_states.append(x)
    
    # Return branches that ended with the chosen branching condition, end states that involved the branching condition, and the chosen variable value for each child of a branch that ended with the chosen branching condition.
    return (relevant_branches, relevant_end_states, relevant_branch_children, child_branch_values)


def get_branch_variable_history(branchid, var_chosen):
    # Get the history of a given ADAPT variable value for a given branch up to the parent.
    heritage = adaptlibs.get_branch_heritage(branchid)
    relevant_branches_set = ', '.join([repr(int(x)) for x in heritage])
    branch_var_inputs = adaptlibs.GetDataAccessObject().branch_input_blobsSelectAllWhereInBranchIDandState(relevant_branches_set=relevant_branches_set)
    
    branch_values = {}
    for x in heritage:
        for y in branch_var_inputs:
            if int(x) == int(y[0]):
                branch_var_input = [k for k in y]
                branch_var_input_data = pickle.loads(branch_var_input[3])
                branch_var_value = branch_var_input_data[1][var_chosen]
                branch_values[int(y[0])] = branch_var_value
    
    # Return the inputs, the heritage of the chosen branch, and the values for each branch in the heritage.
    return (branchid, var_chosen, heritage, branch_values)


# Get the mean value time series of a plot variable over an experiment.
def get_exp_plot_variable_mean_value(experimentid, plotfile, plot_values, num_bins, binscheme, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    use_processors = kwargs.get('use_processors', 1)
    # num_bins refers to the number of bins to divide the time space into.
    # binscheme refers to the way to divide the time space. Currently the only option is 'lin'. 'log' will be added at some point.
    
    # Gather the data. This can take a long time.
    if type(plot_values) is None:
        plot_values = adaptlibs.get_exp_heritage_plot_data(experimentid, plotfile, use_old_data=use_old_data, use_processors=use_processors)
    end_states_and_probs = adaptlibs.GetDataAccessObject().branchesSelectEndStatesProbability(experimentid=experimentid)
    end_state_probs = {}
    for end in end_states_and_probs:
        end_state_probs[end[0]] = end[1]
    
    # Identify the time span of the longest sequence.
    branch_with_longest_time_span = 0
    longest_time_span = 0.0
    for end in plot_values:
        if abs(plot_values[end][-1][0] - plot_values[end][0][0]) > longest_time_span:
            min_time = plot_values[end][0][0]
            max_time = plot_values[end][-1][0]
            longest_time_span = abs(plot_values[end][-1][0] - plot_values[end][0][0])
            branch_with_longest_time_span = end
    
    if binscheme == 'lin':
        mean_time_vector = []
        mean_time_vector.append(min_time)
        last = min_time
        span = longest_time_span / num_bins
        while last < max_time:
            last = last + span
            if last > max_time:
                pass
            else:
                mean_time_vector.append(last)
    elif binscheme == 'log':
        import math
        mean_time_vector = []
        mean_time_vector.append(min_time)
        log_min_time = math.log10(max(min_time, 1E-20))
        log_max_time = math.log10(max(max_time, 1E-20))
        log_longest_time_span = log_max_time - log_min_time
        last = log_min_time
        span = log_longest_time_span / num_bins
        while last < log_max_time:
            last = last + span
            if last > log_max_time:
                pass
            else:
                mean_time_vector.append(10**last)
        if abs(mean_time_vector[-1] - max_time) > 1.0:
            mean_time_vector.append(max_time)
    
    # Initialize the bins.
    time_bins = {}
    for x, time in enumerate(mean_time_vector):
        if x == len(mean_time_vector) - 1:
            break
        time_bins[x] = {}
        time_bins[x]['min'] = time
        time_bins[x]['max'] = mean_time_vector[x + 1]
        time_bins[x]['member_sequence'] = []
        time_bins[x]['member_value'] = []
    
    # Assign each time step to a bin.
    for end in plot_values:
        for bin in time_bins:
            end_steps = [y for y in plot_values[end] if (y[0] >= time_bins[bin]['min']) and (y[0] < time_bins[bin]['max'])]
            if len(end_steps) == 0:
                pass
            else:
                # Find the time-averaged value of the time steps for this end state in this bin.
                time_average_val = 0.0
                time_average_time = 0.0
                for x, step in enumerate(end_steps):
                    if x == len(end_steps) - 1:
                        break
                    time_average_val += step[1] * (end_steps[x + 1][0] - step[0])
                    time_average_time += (end_steps[x + 1][0] - step[0])
                try:
                    weighted_ave = time_average_val / time_average_time
                except:
                    weighted_ave = None
                if weighted_ave is None:
                    pass
                else:
                    # All time steps in this bin for any particular sequence are collapsed to a single value to represent that sequence.
                    time_bins[bin]['member_sequence'].append(end)
                    time_bins[bin]['member_value'].append(weighted_ave)
    
    # Find the total probability in each bin as well as the weighted-average value.
    totalprob = {}
    prob_times_val = {}
    bin_average = {}
    mark_to_delete = []
    for bin in time_bins:
        totalprob[bin] = 0.0
        prob_times_val[bin] = 0.0
        bin_average[bin] = 0.0
        time_bins[bin]['combined_list'] = [list(a) for a in zip(time_bins[bin]['member_value'], time_bins[bin]['member_sequence'])]
        for val in time_bins[bin]['combined_list']:
            try:
                totalprob[bin] += end_state_probs[val[1]]
                prob_times_val[bin] += end_state_probs[val[1]] * val[0]
            except:
                pass
        if totalprob[bin] == 0.0:
            mark_to_delete.append(bin)
        else:
            bin_average[bin] = prob_times_val[bin] / totalprob[bin]
    
    for bin in mark_to_delete:
        del bin_average[bin]
        del time_bins[bin]
    
    mean_plotting_time = []
    mean_plotting_val = []
    for x, bin in enumerate(time_bins):
        #mean_plotting_time.append(time_bins[bin]['min'])
        #mean_plotting_val.append(mean[bin])
        # Midpoint only.
        mean_plotting_time.append((time_bins[bin]['min'] + time_bins[bin]['max']) / 2)
        mean_plotting_val.append(bin_average[bin])
        #if x == (len(time_bins) - 1):
        #    mean_plotting_time.append(time_bins[bin]['max'])
        #    mean_plotting_val.append(mean[bin])
    return (time_bins, bin_average, mean_plotting_time, mean_plotting_val)


# Get the median value time series of a plot variable over an experiment.
def get_exp_plot_variable_median_value(experimentid, plotfile, plot_values, num_bins, binscheme, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    use_processors = kwargs.get('use_processors', 1)
    # num_bins refers to the number of bins to divide the time space into.
    # binscheme refers to the way to divide the time space. Currently the only option is 'lin'. 'log' will be added at some point.
    
    # Gather the data. This can take a long time.
    if type(plot_values) is None:
        plot_values = adaptlibs.get_exp_heritage_plot_data(experimentid, plotfile, use_old_data=use_old_data, use_processors=use_processors)
    
    end_states_and_probs = adaptlibs.GetDataAccessObject().branchesSelectEndStatesProbability(experimentid=experimentid)
    end_state_probs = {end_state:prob for (end_state, prob) in end_states_and_probs}
    
    # Since we may be using old plot data, some of the end states in the plot data file may not be end states anymore. Clear any that do not appear in end_state_probs to avoid trouble.
    for plot in plot_values.keys():
        if plot not in end_state_probs:
            del plot_values[plot]
    
    # Identify the time span of the longest sequence.
    longest_time_span = 0.0
    for end in plot_values:
        if abs(plot_values[end][-1][0] - plot_values[end][0][0]) > longest_time_span:
            min_time = plot_values[end][0][0]
            max_time = plot_values[end][-1][0]
            longest_time_span = abs(max_time - min_time)
            branch_with_longest_time_span = end
    
    if binscheme == 'lin':
        mean_time_vector = []
        mean_time_vector.append(min_time)
        last = min_time
        span = longest_time_span / num_bins
        while last < max_time:
            last = last + span
            if last > max_time:
                pass
            else:
                mean_time_vector.append(last)
    elif binscheme == 'log':
        import math
        mean_time_vector = []
        mean_time_vector.append(min_time)
        log_min_time = math.log10(max(min_time, 1E-20))
        log_max_time = math.log10(max(max_time, 1E-20))
        log_longest_time_span = log_max_time - log_min_time
        last = log_min_time
        span = log_longest_time_span / num_bins
        while last < log_max_time:
            last = last + span
            if last > log_max_time:
                pass
            else:
                mean_time_vector.append(10**last)
        if abs(mean_time_vector[-1] - max_time) > 1.0:
            mean_time_vector.append(max_time)
    
    # Initialize the bins.
    time_bins = {}
    for x, time in enumerate(mean_time_vector):
        if x == len(mean_time_vector) - 1:
            break
        time_bins[x] = {}
        time_bins[x]['min'] = time
        time_bins[x]['max'] = mean_time_vector[x + 1]
        time_bins[x]['member_sequence'] = []
        time_bins[x]['member_value'] = []
        time_bins[x]['member_time'] = []
        time_bins[x]['synthetic_members'] = []
        time_bins[x]['member_allplot'] = {}
    
    # Assign each time step to a bin.
    for end in plot_values:
        for bin in time_bins:
            end_steps = [x for x in plot_values[end] if (x[0] >= time_bins[bin]['min']) and (x[0] < time_bins[bin]['max'])]
            if len(end_steps) == 0:
                # If a sequence appears in a previous bin but not this one, we create a point for it to move its probability forward. Its time is the midpoint of this bin and its value is either 1E20 or -1E20. The value depends on whether it was in the upper 50th or lower 50th in the previous bin.
                if (end in time_bins[bin - 1]['member_sequence']) or (end in time_bins[bin - 1]['synthetic_members']):
                    time_bins[bin]['synthetic_members'].append(end)
            else:
                # Find the midpoint value of the time steps for this end state in this bin.
                min_time_step = end_steps[0][0]
                max_time_step = end_steps[-1][0]
                mid_time_step = (min_time_step + max_time_step) / 2
                dist_step = []
                for step in end_steps:
                    dist_step.append(abs(step[0] - mid_time_step))
                min_dist_step, min_dist_step_index = min((val, idx) for (idx, val) in enumerate(dist_step))
                time_bins[bin]['member_sequence'].append(end)
                time_bins[bin]['member_time'].append(end_steps[min_dist_step_index][0])
                time_bins[bin]['member_value'].append(end_steps[min_dist_step_index][1])
                time_bins[bin]['member_allplot'][end] = end_steps
    
    # Find the total probability in each bin.
    totalprob = {}
    for bin in time_bins:
        totalprob[bin] = 0.0
        time_bins[bin]['combined_list'] = [list(a) for a in zip(time_bins[bin]['member_value'], time_bins[bin]['member_sequence'], time_bins[bin]['member_time'])]
        for val in time_bins[bin]['combined_list']:
            try:
                totalprob[bin] += end_state_probs[val[1]]
            except:
                pass
        for val2 in time_bins[bin]['synthetic_members']:
            time_bins[bin]['combined_list'].append([None, val2])
            try:
                totalprob[bin] += end_state_probs[val2]
            except:
                pass
    
    # Find the 50th percentile value of each bin.
    prob = {}
    median = {}
    median_time = {}
    bin_finished = {}
    bin_percentiles = {}
    bin_synthetic = {}
    for bin in time_bins:
        bin_percentiles[bin] = {}
        bin_finished[bin] = False
        prob[bin] = 0.0
        median[bin] = None
        bin_synthetic[bin] = {}
        bin_synthetic[bin]['pos'] = 0.0
        bin_synthetic[bin]['neg'] = 0.0
        for x, val in enumerate(time_bins[bin]['combined_list']):
            if val[0] is None:
                # Assign synthetic sequences either 1E20 or -1E20 depending on whether they fell into the high or low half of the last bin.
                if bin_percentiles[bin - 1][val[1]] < 0.5:
                    time_bins[bin]['combined_list'][x][0] = -1E20
                    bin_synthetic[bin]['pos'] += end_state_probs[val[1]] / totalprob[bin]
                else:
                    time_bins[bin]['combined_list'][x][0] = 1E20
                    bin_synthetic[bin]['neg'] += end_state_probs[val[1]] / totalprob[bin]
            prob[bin] += end_state_probs[val[1]]
            bin_percentiles[bin][val[1]] = prob[bin] / totalprob[bin]
            if (bin_finished[bin] is False) and ((prob[bin] / totalprob[bin]) > 0.5):
                try:
                    median[bin] = time_bins[bin]['member_allplot'][val[1]]
                    bin_finished[bin] = True
                except:
                    bin_finished[bin] = True
    
    # If 50% of either the top half or the bottom half of the distribution go away permanently, throw out subsequent medians. Work backwards from the last bin. Stop when we get to one that doesn't get thrown out.
    if int(sys.version[0] == 2):
        for bin in reversed(median.keys()):
            if bin_synthetic[bin]['pos'] >= 0.5:
                del median[bin]
                del time_bins[bin]
            elif bin_synthetic[bin]['neg'] >= 0.5:
                del median[bin]
                del time_bins[bin]
            else:
                break
    else:
        for bin in reversed(list(median.keys())):
            if bin_synthetic[bin]['pos'] >= 0.5:
                del median[bin]
                del time_bins[bin]
            elif bin_synthetic[bin]['neg'] >= 0.5:
                del median[bin]
                del time_bins[bin]
            else:
                break
    
    # Replace synthetic medians that are bookended by real medians with the average of the real medians. This would be done by the plotting tool connecting the points anyway.
    median_gaps = {}
    for bin in median:
        if median[bin] is not None:
            if (max(max(median[bin])) > 1E19) or (min(min(median[bin])) < -1E19):
                median_gaps[bin] = {}
                for previousbin in reversed(range(0, bin)):
                    if not(median[previousbin] > 1E19) and not(median[previousbin] < -1E19):
                        median_gaps[bin]['previousreal'] = previousbin
                        median_gaps[bin]['previousrealval'] = median[previousbin][-1][1]
                        break
                for nextbin in range(bin, len(median) - 1):
                    if not(median[nextbin] > 1E19) and not(median[nextbin] < -1E19):
                        median_gaps[bin]['nextreal'] = nextbin
                        median_gaps[bin]['nextrealval'] = median[nextbin][0][1]
                        break
    
    for bin in median_gaps:
        try:
            median[bin] = [((time_bins[x]['max'] + time_bins[x]['min']) / 2.0) ,((median_gaps[bin]['previousrealval'] + median_gaps[bin]['nextrealval']) / 2.0)]
        except:
            pass
    
    median_plotting_time = []
    median_plotting_val = []
    time_bins_kill = []
    for x, bin in enumerate(time_bins):
        if (median[bin] is not None) and not(max(max(median[bin])) > 1E19) and not(min(min(median[bin])) < -1E19):
            median_plotting_time.extend([x[0] for x in median[bin]])
            median_plotting_val.extend([x[1] for x in median[bin]])
        else:
            time_bins_kill.append(bin)
    
    for bin in time_bins_kill:
        del time_bins[bin]
        del median[bin]
    
    return (time_bins, median_plotting_time, median_plotting_val, median)


def get_exp_variable_history(expid, var_chosen, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    # Get the history of a given ADAPT variable value for all end states in a given experiment up to the parent.
    (expid, root_branch, end_states, exp_branch_heritage) = adaptlibs.get_exp_branch_heritage(expid, use_old_data=use_old_data)
    branch_var_inputs = adaptlibs.GetDataAccessObject().branch_input_blobsSelectAllWhereInExpIDandState(expid=expid)
    
    branch_values = {}
    for end in end_states:
        branch_values[end] = {}
        for x in exp_branch_heritage[end]:
            for y in branch_var_inputs:
                if int(x) == int(y[0]):
                    branch_var_input = [k for k in y]
                    branch_var_input_data = pickle.loads(branch_var_input[3])
                    branch_var_value = branch_var_input_data[1][var_chosen]
                    branch_values[end][int(y[0])] = branch_var_value
    
    # Return the inputs, the heritage of the chosen experiment, and the string of values for each end state.
    return (expid, var_chosen, end_states, exp_branch_heritage, branch_values)


def job_remove(jid, commit=False):
    rows = adaptlibs.GetDataAccessObject().jobsSelectAllWhereJobID(jobid=jid)
    ssh = AdaptConfig.get_ssh_ssh()
    for row in rows:
        id, branchid, state, hostname, rundir, starttime, endtime, pid = row
        print("cleanup of %s:%s" % (hostname, rundir))
        cmd = '%s %s rm -fr %s' % (ssh, hostname, shell_escape(rundir))
        print('\-->  ', cmd)
        
        command_done = False
        tries = 0
        while (command_done is False) and (tries < 5):
            try:
                rc = subprocess.call(cmd, shell=True)
                tries += 1
                command_done = True
            except:
                tries += 1
                print('Error running %s. Retrying.' % (cmd))
                #sys.stderr.write("ERROR: running %s\n" % (cmd))
                #raise RuntimeError('ERROR: running %s' % (cmd))
        if (command_done is False) and (tries >= 5):
            print('Tried and failed 5 times to run %s.' % (cmd))
    if commit:
        adaptlibs.GetDataAccessObject().job_output_blobsDeleteAllWhere(jobid=jid)
        adaptlibs.GetDataAccessObject().job_output_infoDeleteAllWhere(jobid=jid)
        adaptlibs.GetDataAccessObject().job_checkpoints_takenDeleteAllWhere(jobid=jid)
        adaptlibs.GetDataAccessObject().job_output_lsDeleteAllWhere(jobid=jid)
        adaptlibs.GetDataAccessObject().jobsDeleteAllWhere(jobid=jid)


def children_identify(bid):
    # Get branches that use the input branch as a parent.
    child_branches = adaptlibs.GetDataAccessObject().branchesSelectIDWhereParentID(parentid=bid)
    children = [0] * len(child_branches)
    for x in range(0, len(child_branches)):
        children[x] = child_branches[x][0]
    return children


def exp_children_identify(expid):
    # Get branch relationships and identify root branch.
    ids_and_parentids = adaptlibs.GetDataAccessObject().branchesSelectIDParentIDWhereExpID(expid=expid)
    try:
        root_branch = [x[0] for x in ids_and_parentids if (x[1] is None)][0]
    except IndexError:
        print('There appear to be no branches under experimentid %s.' % str(expid))
    # Make a dictionary with key parent and value list of children.
    branch_children = {}
    for line in ids_and_parentids:
        if line[1] not in branch_children:
            branch_children[line[1]] = []
        branch_children[line[1]].append(line[0])
    return branch_children


def identify_resulting_end_states(bid, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    # Get all end states that use the input branch.
    branch = adaptlibs.GetDataAccessObject().branchesSelectAllWhereBranchID(branchid=bid)
    expid = branch[0][1]
    (expid, root_branch, end_branches, exp_branch_heritage) = adaptlibs.get_exp_branch_heritage(expid, use_old_data=use_old_data)
    
    resulting_end_states = []
    for end in end_branches:
        if bid in exp_branch_heritage[end]:
            resulting_end_states.append(end)
    
    return resulting_end_states


def get_end_branches(expid):
    # Given an experimentid, this function finds all end branches. A branch is an end branch if no branches count it as a parent.
    end_branches = adaptlibs.GetDataAccessObject().branchesSelectEndStates(expid=expid)
    return end_branches


def get_end_state_progress(expid):
    # Given an experiment, get the simtime values for each end state. This can be used to track the progress of an experiment. Only jobs that finished and led to no child branches are included, so while the experiment is running this may be substantially less than the total number of end states.
    jobid_times = adaptlibs.GetDataAccessObject().job_output_infoSelectEndStateTimes(expid=expid)
    jobid_times = list(set(jobid_times))
    return jobid_times


def histogram_end_state_progress(expid, num_bins=None):
    # Given an experiment, plot the progress of finished end states in linear and log bins. Requires matplotlib and numpy (only if num_bins is not specified).
    try:
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import matplotlib.ticker as mtick
    except ImportError:
        print('adaptlibs.histogram_end_state_progress requires matplotlib, which was not found.')
        return
    
    jobid_times = adaptlibs.get_end_state_progress(expid)
    times_list = [float(x[1]) for x in jobid_times]
    mintime = min(times_list)
    maxtime = max(times_list)
    
    # Use the Freedman-Diaconis rule to determine the number of bins if one was not given.
    if num_bins is None:
        try:
            import numpy as np
        except ImportError:
            print('adaptlibs.histogram_end_state_progress requires numpy if num_bins is not specified. numpy was was not found.')
            return
        q75, q25 = np.percentile(times_list, [75 ,25])
        iqr = q75 - q25
        n_1_3 = len(times_list) ** (1. / 3)
        fd_size = (2 * iqr) / n_1_3
        fd_num = int(np.ceil((maxtime - mintime) / fd_size))
        num_bins = fd_num
    
    thistime = time.asctime().replace(' ', '-').replace(':', '-')
    
    # Sequences plotted linear.
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    
    ax1.hist(times_list, bins=num_bins)
    plt.xlabel('End Time (s)')
    ax1.set_ylabel('Number of End States')
    ax2 = ax1.twinx()
    
    sort_timelist = sorted(times_list)
    time_cdf = []
    for x, point in enumerate(sort_timelist):
        time_cdf.append(1 - (x / float(len(sort_timelist))))
    
    ax2.plot(sort_timelist, time_cdf, color='r', linewidth=0.3)
    ax2.set_ylabel('Complement of Cumulative End States Completed')
    for tl in ax2.get_yticklabels():
        tl.set_color('r')
    
    plt.savefig('histogram_end_state_progress_lin_%s_%s.pdf' % (str(expid), thistime), bbox_inches = 'tight')
    plt.savefig('histogram_end_state_progress_lin_%s_%s.png' % (str(expid), thistime), bbox_inches = 'tight')
    plt.close()
    
    # Sequences plotted log.
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    
    ax1.hist(times_list, bins=num_bins)
    #ax1.set_yscale('log')
    plt.xlabel('End Time (s)')
    plt.yscale('log', nonposy='clip')
    ax1.set_ylabel('Number of End States')
    ax2 = ax1.twinx()
    
    sort_timelist = sorted(times_list)
    time_cdf = []
    for x, point in enumerate(sort_timelist):
        time_cdf.append(1 - (x / float(len(sort_timelist))))
    
    ax2.semilogy(sort_timelist, time_cdf, color='r', linewidth=0.3)
    ax2.set_ylabel('Complement of Cumulative End States Completed')
    for tl in ax2.get_yticklabels():
        tl.set_color('r')
    
    plt.savefig('histogram_end_state_progress_semilogy_%s_%s.pdf' % (str(expid), thistime), bbox_inches = 'tight')
    plt.savefig('histogram_end_state_progress_semilogy_%s_%s.png' % (str(expid), thistime), bbox_inches = 'tight')
    plt.close()
    
    # loglog.
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    
    ax1.hist(times_list, bins=num_bins)
    plt.xlabel('End Time (s)')
    plt.xscale('log', nonposy='clip')
    plt.yscale('log', nonposy='clip')
    ax1.set_ylabel('Number of End States')
    ax2 = ax1.twinx()
    
    sort_timelist = sorted(times_list)
    time_cdf = []
    for x, point in enumerate(sort_timelist):
        time_cdf.append(1 - (x / float(len(sort_timelist))))
    
    ax2.semilogy(sort_timelist, time_cdf, color='r', linewidth=0.3)
    ax2.set_ylabel('Complement of Cumulative End States Completed')
    for tl in ax2.get_yticklabels():
        tl.set_color('r')
    
    plt.savefig('histogram_end_state_progress_loglog_%s_%s.pdf' % (str(expid), thistime), bbox_inches = 'tight')
    plt.savefig('histogram_end_state_progress_loglog_%s_%s.png' % (str(expid), thistime), bbox_inches = 'tight')
    plt.close()


def histogram_last_completed_progress(expid, num_bins=None, **kwargs):
    use_old_data = kwargs.get('use_old_data', False)
    # An expansion of histogram_end_state_progress to include the last job to have finished in every sequence. This will naturally include many branches twice if the DET has not finished.
    try:
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import matplotlib.ticker as mtick
    except ImportError:
        print('adaptlibs.histogram_last_completed_progress requires matplotlib, which was not found.')
        return
    
    # Find all the sequences.
    # Find the last completed job in each sequence.
    # Find the simtime for each of those completed jobs.
    (expid, root_branch, end_branches, exp_branch_heritage) = adaptlibs.get_exp_branch_heritage(expid, use_old_data=use_old_data)
    complete_branches = adaptlibs.GetDataAccessObject().branches_completedSelectAllWhere(expid=expid)
    # Translate complete_branches into a branch-keyed dictionary.
    complete_branches_dict = {}
    for x in complete_branches:
        complete_branches_dict[x[1]] = x[2]
    
    job_completion_times = adaptlibs.GetDataAccessObject().job_output_infoSelectInfoWhereFinishedByExpID(expid=expid)
    # Translate job_completion_times to a job-keyed dictionary.
    job_completion_times_dict = {}
    for x in job_completion_times:
        if 'simtime' in x[1]:
            job_completion_times_dict[x[0]] = float(x[2])
    
    current_times = {}
    for end in end_branches:
        if (end in complete_branches_dict) and (complete_branches_dict[end] in job_completion_times_dict):
            current_times[end] = job_completion_times_dict[complete_branches_dict[end]]
        else:
            # Find position of the end state in its own heritage (should usually be -1, but stuff happens) and pick the branch one before it.
            current_times[end] = job_completion_times_dict[complete_branches_dict[exp_branch_heritage[end][exp_branch_heritage[end].index(end) - 1]]]
    
    times_list = [float(current_times[x]) for x in current_times]
    mintime = min(times_list)
    maxtime = max(times_list)
    
    # Use the Freedman-Diaconis rule to determine the number of bins if one was not given.
    if num_bins is None:
        try:
            import numpy as np
        except ImportError:
            print('adaptlibs.histogram_end_state_progress requires numpy if num_bins is not specified. numpy was was not found.')
            return
        q75, q25 = np.percentile(times_list, [75 ,25])
        iqr = q75 - q25
        n_1_3 = len(times_list) ** (1. / 3)
        fd_size = (2 * iqr) / n_1_3
        fd_num = int(np.ceil((maxtime - mintime) / fd_size))
        num_bins = fd_num
    
    thistime = time.asctime().replace(' ', '-').replace(':', '-')
    
    plt.hist(times_list, bins=num_bins)
    plt.yscale('log', nonposy='clip')
    plt.xlabel('End Time (s)')
    plt.ylabel('Number of End States')
    plt.savefig('histogram_last_completed_progress_%s_%s.pdf' % (str(expid), thistime), bbox_inches = 'tight')
    plt.savefig('histogram_last_completed_progress_%s_%s.png' % (str(expid), thistime), bbox_inches = 'tight')
    plt.close()


def get_branching_condition_frequency(expid):
    # Tells how many times each branching condition is used in expid.
    jobs_branching = adaptlibs.GetDataAccessObject().job_output_infoSelectInfoWhereKAndExpID(k='code', expid=expid)
    all_bcs = [x[2] for x in jobs_branching]
    unique_set = set(all_bcs)
    numbers = {x: all_bcs.count(x) for x in unique_set}
    return (expid, numbers)


def dyi_match_timesteps(timestep_iterable):
    end_state = timestep_iterable[0]
    end_state_final_consequence_state = timestep_iterable[1]
    ref_timesteps = timestep_iterable[2]
    timestep_adjusted_vals = {}
    for step in end_state_final_consequence_state:
        min_dist = 1.0E20
        min_step = 0
        for refstep in ref_timesteps:
            if abs(step[0] - refstep) < min_dist:
                min_dist = abs(step[0] - refstep)
                min_step = refstep
        timestep_adjusted_vals[min_step] = [min_step, step[1]]
    print('Finished matching timesteps for %s.' % (str(end_state)))
    return (end_state, timestep_adjusted_vals)


def dynamic_importance_measure_calculation(expid, bc_chosen, var_chosen, DYI_name, DYI_type, consequence_file, var_value_nonoccurrence, consequence_type, **kwargs):
    scaled = kwargs.get('scaled', False)
    scaled_mintime = kwargs.get('scaled_mintime', -1.0E20)
    use_old_data = kwargs.get('use_old_data', False)
    use_processors = kwargs.get('use_processors', 1)
    
    # (expid, bc_chosen, var_chosen, DYI_name, DYI_type, consequence_file, var_value_nonoccurrence, DYI_value, pickled_value) = adaptlibs.dynamic_importance_measure_calculation(expid=2, bc_chosen=90003, var_chosen='V1601', DYI_name='start time', DYI_type=2, consequence_file='MELPTF_COR-DMH2-TOT', var_value_nonoccurrence=1.0)
    
    # This will calculate the value of a chosen type of dynamic importance measure (DYI) across the end states of an experiment.
    # Types are:
    # DYI1: R(x=1)/R(x=0), the ratio of expected consequence of occurrence to consequence of non-occurrence, a point measure
    # DYI2: R(x%ile)/R(x=0), the ratio of consequence of each uncertain sample against the consequence of non-occurrence
    # DYI3: R(x%ile)/R(x=1), the ratio of consequence of each uncertain sample against the expected consequence of occurrence
    
    # Get end states relevant to the chosen branching condition and variable.
    (relevant_branches, relevant_end_states, relevant_branch_children, child_branch_values) = adaptlibs.find_var_relevant_branches_and_end_states(expid, bc_chosen, var_chosen, use_old_data=use_old_data)
    
    # Get the probability of each relevant end state.
    end_state_prob = {}
    relevantbranches_sql = ', '.join([str(x) for x in relevant_end_states])
    all_branches = adaptlibs.GetDataAccessObject().branchesSelectProbWhereInBranchID(relevant_branches_set=relevantbranches_sql)
    end_state_prob = {x[0]: float(x[1]) for x in all_branches}
    
    # Get the total probability of the relevant end states.
    end_state_prob_sum = sum(end_state_prob.values())
    
    # Get branching values.
    (experimentid, root_branch, end_states, heritage) = adaptlibs.get_exp_branch_heritage(expid, use_old_data=use_old_data)
    raw_branch_var_inputs = adaptlibs.GetDataAccessObject().branch_input_blobsSelectAllWhereInBranchIDandState(relevant_branches_set=relevantbranches_sql)
    
    branch_var_inputs = {x[0]: pickle.loads(x[3])[1] for x in raw_branch_var_inputs}
    branch_var_value = {x: branch_var_inputs[x][var_chosen] for x in branch_var_inputs}
    
    # Find the unique values of the variable.
    var_values = list(set(branch_var_value.values()))
    
    # Find consequences.
    end_state_consequence_plots_all = adaptlibs.get_exp_heritage_plot_data(expid, consequence_file, use_old_data=use_old_data, use_processors=use_processors)
    # Reduce to only relevant end_states.
    end_state_consequence_plots = {}
    for end in end_state_consequence_plots_all:
        if end in relevant_end_states:
            end_state_consequence_plots[end] = end_state_consequence_plots_all[end]
    
    # Take either the final or the peak value of the consequence measure from each end state.
    end_state_final_consequence = {}
    if consequence_type == 'LAST':
        # Take the final value of the consequence measure from each end state.
        for end_state in end_state_consequence_plots:
            if end_state_consequence_plots[end_state][-1][0] > scaled_mintime:
                end_state_final_consequence[end_state] = [end_state_consequence_plots[end_state][-1]]
    elif consequence_type == 'PEAK':
        # Take the peak value of the consequence measure from each end state.
        for end_state in end_state_consequence_plots:
            if sorted(end_state_consequence_plots[end_state])[-1][0] > scaled_mintime:
                last_peak = -1.0E20
                for step in end_state_consequence_plots[end_state]:
                    if step[1] > last_peak:
                        last_peak = step[1]
                        end_state_final_consequence[end_state] = [step]
    elif consequence_type == 'VALLEY' or consequence_type == 'GLENN':
        # Take the lowest value of the consequence measure from each end state.
        for end_state in end_state_consequence_plots:
            if sorted(end_state_consequence_plots[end_state])[-1][0] > scaled_mintime:
                last_valley = 1.0E20
                for step in end_state_consequence_plots[end_state]:
                    if step[1] < last_valley:
                        last_valley = step[1]
                        end_state_final_consequence[end_state] = [step]
    elif consequence_type == 'MIDPOINT':
        # Take the middle value (by time) of the consequence measure from each end state.
        for end_state in end_state_consequence_plots:
            if sorted(end_state_consequence_plots[end_state])[-1][0] > scaled_mintime:
                end_sorted_plot = sorted(end_state_consequence_plots[end_state])
                end_mintime = end_sorted_plot[0][0]
                end_maxtime = end_sorted_plot[-1][0]
                end_midtime = (end_mintime + end_maxtime) / 2.0
                end_distance = [abs(x[0] - end_midtime) for x in end_sorted_plot]
                end_minindex = min(enumerate(end_distance), key=operator.itemgetter(1))[0]
                end_state_final_consequence[end_state] = [end_state_consequence_plots[end_state][end_minindex]]
    elif consequence_type == 'ONEQUARTERTIME':
        # Take the middle value (by time) of the consequence measure from each end state.
        for end_state in end_state_consequence_plots:
            if sorted(end_state_consequence_plots[end_state])[-1][0] > scaled_mintime:
                end_sorted_plot = sorted(end_state_consequence_plots[end_state])
                end_mintime = end_sorted_plot[0][0]
                end_maxtime = end_sorted_plot[-1][0]
                end_onequartertime = (end_mintime + end_maxtime) / 4.0
                end_distance = [abs(x[0] - end_onequartertime) for x in end_sorted_plot]
                end_minindex = min(enumerate(end_distance), key=operator.itemgetter(1))[0]
                end_state_final_consequence[end_state] = [end_state_consequence_plots[end_state][end_minindex]]
    elif consequence_type == 'THREEQUARTERTIME':
        # Take the middle value (by time) of the consequence measure from each end state.
        for end_state in end_state_consequence_plots:
            if sorted(end_state_consequence_plots[end_state])[-1][0] > scaled_mintime:
                end_sorted_plot = sorted(end_state_consequence_plots[end_state])
                end_mintime = end_sorted_plot[0][0]
                end_maxtime = end_sorted_plot[-1][0]
                end_threequartertime = ((end_mintime + end_maxtime) * 3.0) / 4.0
                end_distance = [abs(x[0] - end_threequartertime) for x in end_sorted_plot]
                end_minindex = min(enumerate(end_distance), key=operator.itemgetter(1))[0]
                end_state_final_consequence[end_state] = [end_state_consequence_plots[end_state][end_minindex]]
    elif consequence_type == 'ALL':
        # Take all values of the consequence measure from each end state. Will take a lot longer.
        pass
        for end_state in end_state_consequence_plots:
            end_state_final_consequence[end_state] = end_state_consequence_plots[end_state]
    else:
        print("Unrecognized consequence_type. Currently supports 'ALL', 'LAST', 'PEAK', 'VALLEY', 'ONEQUARTERTIME', 'THREEQUARTERTIME', and 'MIDPOINT'.")
        sys.exit(1)
    
    # Optionally scale consequence values linearly from max to min value across the sequences so that they vary from 0.0 to 1.0.
    if scaled:
        scaled_min = 1.0E20
        scaled_max = -1.0E20
        for end_state in end_state_final_consequence:
            for step in end_state_final_consequence[end_state]:
                current_val = step[1]
                if current_val < scaled_min:
                    scaled_min = current_val
                if current_val > scaled_max:
                    scaled_max = current_val
        # Adjust values to scaling.
        for end_state in end_state_final_consequence:
            for x in range(0, len(end_state_final_consequence[end_state])):
                end_state_final_consequence[end_state][x][1] = (end_state_final_consequence[end_state][x][1] - scaled_min) / (scaled_max - scaled_min)
    
    # At this point, there may not be an entry in end_state_final_consequence for every end state.
    # Plot times between branches are not necessarily equal, may be off by a few fractions of a second.
    branch_with_most_times = 0
    most_time_steps = 0
    for end_state in end_state_final_consequence:
        if len(end_state_final_consequence[end_state]) > most_time_steps:
            branch_with_most_times = end_state
            most_time_steps = len(end_state_final_consequence[end_state])
    
    ref_timesteps = []
    for step in sorted(end_state_final_consequence[branch_with_most_times]):
        ref_timesteps.append(step[0])
    
    # Match each time step of each end state to a step from the end state with the most time steps. This can be done in parallel, distributing end states.
    if use_processors < 2:
        timestep_adjusted_vals = {}
        for end_state in end_state_final_consequence:
            timestep_adjusted_vals[end_state] = {}
            for step in end_state_final_consequence[end_state]:
                # Assign to the closest time step of the sequence with the most time steps.
                timestep_adjusted_vals[end_state][min(ref_timesteps, key=lambda y:abs(y - step[0]))] = [min(ref_timesteps, key=lambda y:abs(y - step[0])), step[1]]
    else:
        # Create a single iterable for the multiprocessing function.
        timestep_match_iterable = []
        for end_state in end_state_final_consequence:
            timestep_match_iterable.append([])
            timestep_match_iterable[-1].append(end_state)
            timestep_match_iterable[-1].append(end_state_final_consequence[end_state])
            timestep_match_iterable[-1].append(ref_timesteps)
        
        from multiprocessing import Pool as ProcessPool
        mypool = ProcessPool(use_processors)
        results = mypool.map(adaptlibs.dyi_match_timesteps, timestep_match_iterable)
        
        timestep_adjusted_vals = {}
        for end_state, adjusted_vals in results:
            timestep_adjusted_vals[end_state] = adjusted_vals
        
        del results
        mypool.close()
        mypool.join()
    
    # If the type of DYI is 1 or 3, calculate the weighted average consequence of all values of occurrence for each reference time step.
    consequence_occurrence_average = {}
    consequence_occurrence_prob = {}
    for step in ref_timesteps:
        consequence_occurrence_average[step] = 0.0
        consequence_occurrence_prob[step] = 0.0
    
    if ((DYI_type == 1) or (DYI_type == 3)):
        for end_state in timestep_adjusted_vals:
            if end_state in branch_var_value and str(branch_var_value[end_state]) != str(var_value_nonoccurrence):
                for step in timestep_adjusted_vals[end_state]:
                    consequence_occurrence_prob[timestep_adjusted_vals[end_state][step][0]] += float(end_state_prob[end_state])
                    consequence_occurrence_average[timestep_adjusted_vals[end_state][step][0]] += timestep_adjusted_vals[end_state][step][1] * float(end_state_prob[end_state])
        for step in ref_timesteps:
            if consequence_occurrence_average[step] == 0.0:
                pass
            else:
                consequence_occurrence_average[step] = consequence_occurrence_average[step] / consequence_occurrence_prob[step]
    
    # If the type of DYI is 1 or 2, calculate the weighted average consequence of all branches with non-occurrence.
    consequence_nonoccurrence_average = {}
    consequence_nonoccurrence_prob = {}
    for step in ref_timesteps:
        consequence_nonoccurrence_average[step] = 0.0
        consequence_nonoccurrence_prob[step] = 0.0
    
    if ((DYI_type == 1) or (DYI_type == 2)):
        for end_state in timestep_adjusted_vals:
            if end_state in branch_var_value and str(branch_var_value[end_state]) == str(var_value_nonoccurrence):
                for step in timestep_adjusted_vals[end_state]:
                    consequence_nonoccurrence_prob[timestep_adjusted_vals[end_state][step][0]] += float(end_state_prob[end_state])
                    consequence_nonoccurrence_average[timestep_adjusted_vals[end_state][step][0]] += timestep_adjusted_vals[end_state][step][1] * float(end_state_prob[end_state])
        for step in ref_timesteps:
            if consequence_nonoccurrence_average[step] == 0.0:
                pass
            else:
                consequence_nonoccurrence_average[step] = consequence_nonoccurrence_average[step] / consequence_nonoccurrence_prob[step]
    
    # If the type of DYI is 2 or 3, calculate the weighted average consequence of all branches with each branching value.
    consequence_occurrence_by_value_average = {}
    consequence_occurrence_by_value_prob = {}
    
    if ((DYI_type == 2) or (DYI_type == 3)):
        for val in var_values:
            if val == var_value_nonoccurrence:
                pass
            else:
                consequence_occurrence_by_value_average[val] = {}
                consequence_occurrence_by_value_prob[val] = {}
                for step in ref_timesteps:
                    consequence_occurrence_by_value_average[val][step] = 0.0
                    consequence_occurrence_by_value_prob[val][step] = 0.0
        for val in var_values:
            if val == var_value_nonoccurrence:
                pass
            else:
                for end_state in timestep_adjusted_vals:
                    if end_state in branch_var_value and str(branch_var_value[end_state]) == str(val):
                        for step in timestep_adjusted_vals[end_state]:
                            consequence_occurrence_by_value_prob[val][step] += float(end_state_prob[end_state])
                            consequence_occurrence_by_value_average[val][step] += timestep_adjusted_vals[end_state][step][1] * float(end_state_prob[end_state])
        for val in var_values:
            if val == var_value_nonoccurrence:
                pass
            else:
                for step in ref_timesteps:
                    if consequence_occurrence_by_value_average[val][step] == 0.0:
                        pass
                    else:
                        consequence_occurrence_by_value_average[val][step] = consequence_occurrence_by_value_average[val][step] / consequence_occurrence_by_value_prob[val][step]
    
    # Assemble the ratios into importance measures.
    DYI_value = {}
    if DYI_type == 1:
        DYI_value[DYI_name] = {}
        DYI_value[DYI_name]['type1'] = {}
        for step in consequence_occurrence_average:
            try:
                DYI_value[DYI_name]['type1'][step] = consequence_occurrence_average[step] / consequence_nonoccurrence_average[step]
            except ZeroDivisionError:
                pass
    if DYI_type == 2:
        DYI_value[DYI_name] = {}
        for value in consequence_occurrence_by_value_average:
            DYI_value[DYI_name][value] = {}
            for step in consequence_occurrence_by_value_average[value]:
                try:
                    DYI_value[DYI_name][value][step] = (consequence_occurrence_by_value_average[value][step] / consequence_nonoccurrence_average[step])
                except ZeroDivisionError:
                    pass
    if DYI_type == 3:
        DYI_value[DYI_name] = {}
        for value in consequence_occurrence_by_value_average:
            DYI_value[DYI_name][value] = {}
            for step in consequence_occurrence_by_value_average[value]:
                try:
                    DYI_value[DYI_name][value][step] = (consequence_occurrence_by_value_average[value][step] / consequence_occurrence_average[step])
                except ZeroDivisionError:
                    pass
    
    # Save a dump of the DYI values.
    pickled_value = pickle.dumps([DYI_value], protocol=0)
    pickle.dump(DYI_value, open((DYI_name + str(DYI_type) + '.pick'), 'wb'))
    
    # Submit the value or set of values to the table.
    #if ((DYI_type == 1) or (DYI_type == 2)):
    #    adaptlibs.GetDataAccessObject().dynamic_importance_measuresInsertAllType12(expid=expid, bc_chosen=bc_chosen, var_chosen=var_chosen, DYI_name=DYI_name, DYI_type=int(DYI_type), consequence_file=consequence_file, var_value_nonoccurrence=var_value_nonoccurrence, value=repr(DYI_value))
    #elif (DYI_type == 3):
    #    adaptlibs.GetDataAccessObject().dynamic_importance_measuresInsertAllType3(expid=expid, bc_chosen=bc_chosen, var_chosen=var_chosen, DYI_name=DYI_name, DYI_type=int(DYI_type), consequence_file=consequence_file, value=repr(DYI_value))
    
    return (expid, bc_chosen, var_chosen, DYI_name, DYI_type, consequence_file, var_value_nonoccurrence, DYI_value, repr(DYI_value))


def pretend_branch_never_completed(bid, commit=False):
    removebids = [bid]
    while True:
        removebids_sql = ', '.join([str(x) for x in removebids])
        rows = adaptlibs.GetDataAccessObject().branchesSelectIDWhereInParentIDAndNotID(relevant_branches_set=removebids_sql)
        if len(rows) == 0:
            break
        for row in rows:
            removebids.append(row[0])
    removebids = list(reversed(sorted(removebids)))
    print(removebids)
    if commit is True:
        for bid2 in removebids:
            adaptlibs.branch_remove(bid2, True)
    elif commit is False:
        for bid2 in removebids:
            adaptlibs.branch_remove(bid2, False)


def simulate_execution_timings(eid, nprocs, print_info=True):
    
    rows = adaptlibs.GetDataAccessObject().branches_jobs_job_output_infoSelectAllWhereBranchesComplete(expid=eid)
    bid_parent = {}
    bid_secs = {}
    for row in rows:
        bid, pid, jid, secs = row
        bid_parent[bid] = pid
        bid_secs[bid] = int(secs)
    
    procs_free = list(range(nprocs))
    procs_running = {} # procid to bid
    bids_to_run = sorted(bid_parent.keys())
    bids_to_procid = {}
    bids_seconds_left = {} # bid to seconds left
    bids_completed = {}
    
    t = 0
    day = 0
    while bids_to_run or procs_running:
        killtorun = []
        for b in bids_to_run:
            if not procs_free:
                break
            elif bid_parent[b] and bid_parent[b] not in bids_completed:
                continue
            procid = procs_free.pop(0)
            procs_running[procid] = b
            bids_to_procid[b] = procid
            bids_seconds_left[b] = bid_secs[b]
            killtorun.append(b)
        for b in killtorun:
            bids_to_run.remove(b)
        killbs = []
        killprocids = []
        
        mingap = max(1, t % 86400)
        for bid in bids_to_procid:
            mingap = min(mingap, bids_seconds_left[bid])
        #print('mingap is', mingap)
        
        t += mingap
        for bid in bids_to_procid:
            bids_seconds_left[bid] -= mingap
            if bids_seconds_left[bid] == 0:
                if print_info:
                    print('at time', t, 'bid', bid, 'finishes', 'procs_saturated', len(procs_running))
                killbs.append(bid)
                killprocids.append(bids_to_procid[bid])
                bids_completed[bid] = 1
        for bid in killbs:
            del bids_to_procid[bid]
            del bids_seconds_left[bid]
        for procid in killprocids:
            del procs_running[procid]
            procs_free.append(procid)
        if t % 86400 == 0:
            if print_info:
                print('after day', day + 1, ',', len(bids_completed), 'branches complete')
                day += 1
    if print_info:
        print('at end time', ',', len(bids_completed), 'branches complete')
    return t


def simulate_execution_timings_serial_parallel(eid, nprocs):
    return (simulate_execution_timings(eid, 1, False),
            simulate_execution_timings(eid, nprocs, False))


def get_branch_rundir(id):
    rows = adaptlibs.GetDataAccessObject().jobsSelectAllWhereBranch(branchid=str(id))
    rundir = None
    for row in rows:
        rundir = row['rundir']
    return rundir


def get_branch_hostname(id):
    rows = adaptlibs.GetDataAccessObject().jobsSelectAllWhereBranch(branchid=str(id))
    hostname = None
    for row in rows:
        hostname = row['hostname']
    return hostname


def get_bid_from_jid(jid):
    rows = adaptlibs.GetDataAccessObject().jobsSelectAllWhereJobID(jobid=str(jid))
    bid = None
    for row in rows:
        bid = row[1]
    return bid


def get_branch_scripts(sim):
    rows = adaptlibs.GetDataAccessObject().branch_scriptsSelectAllWhereSimname(simname=sim)
    script = None
    for row in rows:
        script = row[1]
    return script


def run_script_for_branch(id, script, args):
    ssh = AdaptConfig.get_ssh_ssh()
    scp = AdaptConfig.get_ssh_scp()
    var = None
    arg = ""
    
    for val in args:
        arg += val + " "
    
    script_file = get_branch_script_file(script)
    loc = AdaptConfig.get_script_loc()
    rundir = get_branch_rundir(id)
    hostname = get_branch_hostname(id)
    if adaptlibs.get_platform() == 'windows':
        print('A Windows-incompatible command has been encountered on Windows. Expect errors.')
    else:
        cmd = scp + " " + loc + "/" + script_file + " " + hostname + ":" + rundir
    subprocess.call(cmd, shell=True)
    
    if adaptlibs.get_platform() == 'windows':
        print('A Windows-incompatible command has been encountered on Windows. Expect errors.')
    else:
        cmd = ssh + " " + hostname + " 'cd " + rundir + ";" + rundir + "/" + script_file + " " + arg + "'"
    print(cmd)
    subprocess.call(cmd, shell=True)
    
    file = "adapt_script_out_" + str(id) + ".txt"
    if adaptlibs.get_platform() == 'windows':
        print('A Windows-incompatible command has been encountered on Windows. Expect errors.')
    else:
        cmd = scp + " " + hostname + ":" + rundir + "/adapt_script_out.txt " + file
    subprocess.call(cmd, shell=True)
    
    lines = open(file).readlines()
    
    for line in lines:
        if len(line.split()) != 1:
            return None
        else:
            var = line.split()[0]
            break
    
    if adaptlibs.get_platform() == 'windows':
        print('A Windows-incompatible command has been encountered on Windows. Expect errors.')
    else:
        cmd = "rm -fr " + file
    subprocess.call(cmd, shell=True)
    
    if adaptlibs.get_platform() == 'windows':
        print('A Windows-incompatible command has been encountered on Windows. Expect errors.')
    else:
        cmd = ssh + " " + hostname + " 'rm -fr " + rundir + "/" + script_file + "'"
    subprocess.call(cmd, shell=True)
    
    if adaptlibs.get_platform() == 'windows':
        print('A Windows-incompatible command has been encountered on Windows. Expect errors.')
    else:
        cmd = ssh + " " + hostname + " 'rm -fr " + rundir + "/adapt_script_out.txt'"
    subprocess.call(cmd, shell=True)
    
    return var


def get_branch_script_file(name):
    rows = adaptlibs.GetDataAccessObject().branch_scriptsSelectAllWhereScriptname(scriptname=name)
    script = None
    for row in rows:
        script = row[2]
    return script


class CommandThread(threading.Thread):
    def __init__(self, cmd):
        threading.Thread.__init__(self)
        self.cmd = cmd
        self.rc = None
    
    def run(self):
        self.rc = subprocess.call(self.cmd, shell=True)
        if not os.WIFEXITED(self.rc):
            self.rc = 1
        else:
            self.rc = os.WEXITSTATUS(self.rc)
    
    def is_done(self):
        return self.rc is not None
    
    def get_return_code(self):
        return self.rc


class ssh_executor:
    def __init__(self,
                 hostname=None, directory=None,
                 resource_uid=None, debug=False):
        self.ssh = AdaptConfig.get_ssh_ssh()
        self.scp = AdaptConfig.get_ssh_scp()
        self.hostname = hostname or ''
        self.directory = directory
        self.jobstarted = False
        self.jobaborted = False
        self.commandthread = None
        self.jobcmd = None
        self.resource_uid = resource_uid
        self.debug = debug
    
    def __repr__(self):
        hn = self.hostname
        if not hn.split('.')[0].isdigit():
            hn = hn.split('.')[0]
        return '<ssh:%s>' % (hn)
    
    def _run(self, cmd):
        if self.debug:
            print('+%s' % cmd)
        
        command_done = False
        tries = 0
        while (command_done is False) and (tries < 5):
            rc = subprocess.call(cmd, shell=True)
            tries += 1
            if rc == 0:
                command_done = True
            else:
                print('Error running %s. Retrying.' % (cmd))
        if (command_done is False) and (tries >= 5):
            print('Tried and failed 5 times to run %s.' % (cmd))
            sys.stderr.write("ERROR: running %s\n" % (cmd))
            raise RuntimeError('ERROR: running %s' % (cmd))
    
    def get_hostname(self):
        return self.hostname
    
    def get_directory(self):
        return self.directory
    
    def get_resource_uid(self):
        return self.resource_uid
    
    def make_remote_container(self, name=None):
        if not name:
            d = AdaptConfig.get_server_adaptemp()
            if adaptlibs.get_platform() == 'windows':
                print('A Windows-incompatible command has been encountered on Windows. Expect errors.')
            else:
                cmd = ('%s %s mktemp -d %s') % (self.ssh, self.hostname, os.path.join(d, 'adapt.XXXXXX'))
            
            f = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
            out = f.communicate()[0].strip()
            rc = f.returncode
            if rc:
                raise RuntimeError('ERROR: calling %s.' % (cmd))
            if int(sys.version[0]) == 2:
                pass
            elif int(sys.version[0]) > 2:
                if (not (type(out) == str)):
                    out = out.decode('utf-8')
                else:
                    pass
            else:
                raise ValueError('Unknown python version in adaptlibs.py.')
            return out
        else:
            tmpdir = self.make_remote_container()
            script = """
import os, os.path
d = '%s'
if not os.path.exists(d):
    try:
        os.makedirs(d)
    except Exception:
        if os.access(d,os.F_OK):
            pass
        else:
            raise
""" % (name)
            (tempfd, tempfn) = tempfile.mkstemp(dir=AdaptConfig.get_server_adaptemp(), suffix=".py")
            os.close(tempfd)
            open(tempfn, 'w').write(script)
            if AdaptConfig.get_ssh_shared_dirs():
                # If we use a shared file system, copy directly.
                shutil.copy2(tempfn, tmpdir)
            else:
                cmd = '%s %s %s:%s' % (self.scp, tempfn, self.hostname, tmpdir)
                self._run(cmd)
            os.remove(tempfn)
            cmd = '%s %s python %s' % (self.ssh, self.hostname, os.path.join(tmpdir, os.path.basename(tempfn)))
            self._run(cmd)
            self.cleanup_remote_container(tmpdir)
            if int(sys.version[0]) == 2:
                pass
            elif int(sys.version[0]) > 2:
                if (not (type(name) == str)):
                    name = name.decode('utf-8')
                else:
                    pass
            else:
                raise ValueError('Unknown python version in adaptlibs.py.')
            return name
    
    def make_adaptemp_container(self):
        d = AdaptConfig.get_server_adaptemp()
        out = tempfile.mkdtemp(prefix='adapt.', dir=d)
        return out
    
    def cleanup_remote_container(self, container):
        self._run('%s %s rm -fr %s' % (self.ssh, self.hostname, container))
    
    def copy_files_to_remote_container(self, files, container, make_executable=False):
        if isinstance(files, list):
            file_str = ' '.join([shell_escape(x) for x in files])
        else:
            file_str = shell_escape(files)
        if int(sys.version[0]) == 2 and type(files) is unicode:
            files = files.encode('ascii', 'ignore')
        if AdaptConfig.get_ssh_shared_dirs():
            # If we use a shared file system, copy directly.
            if type(files) is str:
                shutil.copy2(files, container)
            elif type(files) is list:
                for file in files:
                    shutil.copy2(file, container)
            else:
                self._run('%s %s %s:%s' % (self.scp, file_str, self.hostname, container))
        else:
            self._run('%s %s %s:%s' % (self.scp, file_str, self.hostname, container))
        if make_executable:
            if isinstance(files, list):
                file_str = ' '.join([shell_escape(os.path.join(container, os.path.basename(x))) for x in files])
            else:
                file_str = os.path.join(container, shell_escape(os.path.basename(files)))
            self._run('%s %s chmod 755 %s' % (self.ssh, self.hostname, file_str))
    
    def copy_files_to_adaptemp_container(self, files, container, make_executable=False):
        if isinstance(files, list):
            file_str = ' '.join([shell_escape(x) for x in files])
        else:
            file_str = shell_escape(files)
        if int(sys.version[0]) == 2 and type(files) is unicode:
            files = files.encode('ascii', 'ignore')
        if type(files) is str:
            shutil.copy2(files, container)
        elif type(files) is list:
            for file in files:
                shutil.copy2(file, container)
        if make_executable:
            if isinstance(files, list):
                file_str = ' '.join([shell_escape(os.path.join(container, os.path.basename(x))) for x in files])
            else:
                file_str = os.path.join(container, shell_escape(os.path.basename(files)))
            self._run('chmod 755 %s' % (file_str))
    
    def copy_remote_files_to_remote_container(self, sourcehost, sourcefiles, desthost, destcontainer, make_executable=False):
        if isinstance(sourcefiles, list):
            file_str = ' '.join([shell_escape(x) for x in sourcefiles])
        else:
            file_str = shell_escape(sourcefiles)
        if int(sys.version[0]) == 2 and type(file_str) is unicode:
            file_str = file_str.encode('ascii', 'ignore')
        
        self._run('''%s %s:"%s" %s:%s''' % (self.scp, sourcehost, file_str, desthost, destcontainer))
        
        if make_executable:
            if isinstance(sourcefiles, list):
                file_str = ' '.join([shell_escape(os.path.join(destcontainer, os.path.basename(x))) for x in sourcefiles])
            else:
                file_str = os.path.join(destcontainer, shell_escape(os.path.basename(sourcefiles)))
            self._run('%s %s chmod 755 %s' % (self.ssh, self.hostname, file_str))
        return (self.scp, sourcehost, file_str, desthost, destcontainer)
    
    def copy_remote_file_to_local(self, remote, local):
        if AdaptConfig.get_ssh_shared_dirs():
            # If we use a shared file system, copy directly.
            shutil.copy2(shell_escape(shell_escape(remote)), shell_escape(local))
        else:
            self._run('%s %s:%s %s' % (self.scp, self.hostname, shell_escape(shell_escape(remote)), shell_escape(local)))
    
    def copy_remote_file_to_fileobject(self, remote):
        return subprocess.Popen('%s %s cat %s' % (self.ssh, self.hostname, shell_escape(shell_escape(remote))), shell=True, stdout=subprocess.PIPE).stdout
    
    def remote_file_size(self, remote):
        f = subprocess.Popen('%s %s wc -c %s' % (self.ssh, self.hostname, shell_escape(shell_escape(remote))), shell=True, stdout=subprocess.PIPE)
        sz = f.communicate()[0].split()[0]
        print('sz is', sz)
        return int(sz)
    
    def job_start(self, remotecommand):
        if self.debug:
            print('jobcommand is %s' % remotecommand)
        self.jobcmd = remotecommand
        self.commandthread = CommandThread('%s %s "%s"' % (self.ssh, self.hostname, remotecommand))
        self.commandthread.setDaemon(True)
        self.commandthread.start()
        self.jobstarted = True
    
    def job_finished(self):
        return (self.jobstarted and self.commandthread.is_done()) or self.jobaborted
    
    def job_return_code(self):
        if self.commandthread:
            return self.commandthread.get_return_code()
        elif self.jobaborted:
            return -1
        else:
            sys.stderr.write("ERROR: internal error at job_return_code()\n")
            sys.exit(1)
    
    def job_get_command(self):
        return self.jobcmd
    
    def job_mark_as_failed(self):
        self.jobaborted = True


def checkhost(ssh, host): # return 0 on success
    p = subprocess.Popen('%s -oBatchMode=yes %s hostname' % (ssh, host), shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    if not(type(stdout) == str):
        stdout = str(stdout)
    if not(type(stderr) == str):
        stderr = str(stderr)
    
    if len(stdout) > 3:
        # If the hostname is returned in stdout, that means the connection was successful.
        return 0
    elif (len(stderr) > 0) and ('Permission denied' in stderr):
        # Either the remote host asked for a password or the username is bad. The flag -oBatchMode=yes prevents a password prompt from hanging up the check process.
        print('It appears that public key (password-less) authentication is not set up for this user on %s, causing the connection to fail. Error text: "%s". Please correct, close ADAPT, and restart ADAPT.' % (host, stderr))
        return 2
    elif (len(stderr) > 0) and ('Bad owner or permissions' in stderr):
        # It is possible to set incorrect permissions on SSH files such that the connection will be refused.
        print('A file in the SSH configuration for public key (password-less) authentication for this user likely has excessive read/write permissions set, causing the connection to fail. Error text: "%s". Please correct, close ADAPT, and restart ADAPT.' % stderr)
        return 2
    elif (len(stderr) > 0) and ('Name or service not known' in stderr):
        # It is possible to set incorrect permissions on SSH files such that the connection will be refused.
        print('The host %s does not seem to exist on the network. The name may have been mis-spelled in .adaptrc. Error text: "%s". Please correct, close ADAPT, and restart ADAPT.' % (host, stderr))
        return 2
    elif (len(stderr) > 0) and ('No route to host' in stderr):
        # The host is known on the network, but down.
        return -1
    else:
        print('Unable to connect to host %s.' % host)
        return -1


class ssh_resource_manager:
    def __init__(self, debug=False):
        self.debug = debug
        self.ssh = AdaptConfig.get_ssh_ssh()
        print('%s: ADAPT is using the following command for ssh: %s' % (time.asctime(), self.ssh))
        
        self.scp = AdaptConfig.get_ssh_scp()
        print('%s: ADAPT is using the following command for scp: %s' % (time.asctime(), self.scp))
        
        h2 = AdaptConfig.get_ssh_hostlist()
        print('%s: ADAPT is using the following hosts: %s' % (time.asctime(), h2))
        
        self.hosts = []
        self.downhosts = []
        self.directories = {}
        self.host_status = {}
        for h in h2:
            if h in self.host_status:
                if self.host_status[h] == 0:
                    self.hosts.append(h)
                else:
                    pass
            else:
                sys.stdout.write('%s: Checking ADAPT host %s.\n' % (time.asctime(), h))
                sys.stdout.flush()
                status = checkhost(self.ssh, h)
                self.host_status[h] = status
                if status == 0:
                    self.hosts.append(h)
                    sys.stdout.write('%s: ADAPT host %s ready.\n' % (time.asctime(), h))
                    sys.stdout.flush()
                    self.directories[h] = AdaptConfig.get_ssh_directory(h)
                else:
                    sys.stdout.write('%s: ADAPT host %s not ready.\n' % (time.asctime(), h))
                    sys.stdout.flush()
        if self.hosts == []:
            sys.stderr.write("ERROR: Cannot proceed, no hosts available in ssh manager\n")
            sys.exit(1)
        self.hosts_busy = {} # map index into self.hosts -> executor
    
    def get_up_hosts(self):
        return [h for h in list(self.host_status.keys()) if self.host_status[h] == 0]
    
    def get_down_hosts(self):
        return [h for h in list(self.host_status.keys()) if self.host_status[h] != 0]
    
    def has_resources(self):
        for h in self.hosts:
            if h:
                return True
        return False
    
    def get_available_executor(self, needed_host=None):
        idx = 0
        
        #debug
        Ndh = "None"
        if needed_host is not None:
            Ndh = str(needed_host)
            
        #print("needed_host=" + Ndh)
        del Ndh
        #end debug
        
        for h in self.hosts: # host could be None if was disabled
            if(h is None):
                print("Yes, host can be none in a loop.")
            if(h and idx not in self.hosts_busy and not needed_host or needed_host == h):
                try:
                    self.hosts_busy[idx] = ssh_executor(h, self.directories[h], idx, debug=self.debug)
                    result = self.hosts_busy[idx]
                    break
                except:
                    print('Error checking for available executors. This can occur when a host has gone down. adapt-server may require a restart to add the host once it comes back up.')
                    result = None
                    break
            idx += 1
        else:
            result = None
            
        return result
    
    def return_executor(self, executor, disable):
        #print("return_executor is getting hostname")
        host = executor.get_hostname()
        #print("Hostname is " + host)
        
        resource_uid = executor.get_resource_uid()
        print(str(time.asctime()) + ": Deleting self.hosts_busy[" + str(resource_uid) + "]")
        try:
            del self.hosts_busy[resource_uid]
        except:
            pass
        
        if disable:
            sys.stderr.write("WARNING: disabling resource %s/%s\n" % (host, repr(resource_uid)))
            self.hosts[resource_uid] = None
            self.host_status[host] = -1


def get_specific_executor(hostname=None):
    if AdaptConfig.get_server_jobmanager() == 'ssh':
        return ssh_executor(hostname)
    else:
        sys.stderr.write("ERROR: unknown job manager.  Known values are 'ssh'.\n")
        sys.exit(1)


def normalize_branch_probs(expid, method, use_old_data, commit):
    # Normalize the branch conditional probabilities of each branching condition in an experiment to have a summed conditional probability of 1.
    
    # For now, only supporting a proportional method.
    supported_methods = ['PROPORTIONAL']
    if method not in supported_methods:
        print('Unsupported method %s for adaptlibs.normalize_branch_probs. Supported methods: %s' % (method, supported_methods))
        return False
    
    experiment = adaptlibs.GetDataAccessObject().experimentsSelectAllWhere(expid=expid)
    if len(experiment) == 0:
        experiments = adaptlibs.GetDataAccessObject().experimentsSelectAllAll()
        exp_list = []
        for item in experiments:
            exp_list.append(item[0])
        print('Experiment %s does not exist in the database. Currently-existing experiments: %s.' % (str(expid), exp_list))
        return False
    
    # Get all branches.
    all_branches_raw = adaptlibs.GetDataAccessObject().branchesSelectSomeWhereExpid(expid=expid)
    all_branches = {}
    all_branches['experimentid'] = {x[0]: x[1] for x in all_branches_raw}
    all_branches['parentid'] = {x[0]: x[2] for x in all_branches_raw}
    all_branches['branchdescription'] = {x[0]: x[3] for x in all_branches_raw}
    all_branches['probability'] = {x[0]: x[4] for x in all_branches_raw}
    
    # Get a dictionary of all parents and their immediate children.
    branch_children = adaptlibs.exp_children_identify(expid)
    
    # Get heritage for each branch
    (expid, root_branch, end_branches, heritage_all) = adaptlibs.get_exp_branch_heritage(expid, use_old_data=use_old_data)
    
    # Identify when a parent's children do not add to a conditional probability of 1.
    attitude_adjustments = {}
    all_attitude_adjustments = {x: [all_branches['probability'][x], False] for x in all_branches['probability']}
    for parent in branch_children:
        if parent is None or len(branch_children[parent]) == 0:
            # Do nothing.
            pass
        else:
            # Do something.
            parent_prob = all_branches['probability'][parent]
            child_prob_sum = 0.0
            for child in branch_children[parent]:
                child_prob_sum += all_branches['probability'][child]
            # See if an adjustment is necessary.
            if ('PROPORTIONAL' in method) and ((abs(child_prob_sum / parent_prob) < 0.9999) or (abs(child_prob_sum / parent_prob) > 1.0001)):
                adjustment_ratio = parent_prob / child_prob_sum
                for child in branch_children[parent]:
                    # This adjusts immediate children.
                    attitude_adjustments[child] = {}
                    attitude_adjustments[child]['old'] = all_branches['probability'][child]
                    attitude_adjustments[child]['new'] = all_branches['probability'][child] * adjustment_ratio
                    attitude_adjustments[child]['adjustment_ratio'] = adjustment_ratio
                    all_attitude_adjustments[child][0] = all_attitude_adjustments[child][0] * adjustment_ratio
                    all_attitude_adjustments[child][1] = True
                    # Further children
                    for branch in heritage_all:
                        if child in heritage_all[branch] and child != heritage_all[branch][-1]:
                            all_attitude_adjustments[branch][0] = all_attitude_adjustments[branch][0] * adjustment_ratio
                            all_attitude_adjustments[branch][1] = True
    
    if commit is True:
        for branch in all_attitude_adjustments:
            if all_attitude_adjustments[branch][1] is True:
                print('Updating branch %s.' % branch)
                adaptlibs.GetDataAccessObject().branchesUpdateProbabilityWhereBranchID(branchid=branch, probability=('%.8e' % all_attitude_adjustments[branch][0]))
    
    return (all_branches, all_attitude_adjustments)




def change_branch_probs(expid, branching_code, branching_code_descs, branching_code_old_vals_init, branching_code_new_vals_init, commit, **kwargs):
    # Allows for the conditional probabilities of a branching condition to be changed and propagated through an existing experiment.
    # Accepts the results of a previous run of change_branch_probs rather than querying the database for old values. This allows a progressive mutation of the probabilities.
    progressive = kwargs.get('progressive', False)
    new_total_prob = kwargs.get('new_total_prob', None)
    if progressive is True and new_total_prob is None:
        print('Running change_branch_probs in progressive mode requires new_total_prob to be supplied as an input.')
        return None
    
    branching_code_old_vals = [None] * len(branching_code_old_vals_init)
    branching_code_new_vals = [None] * len(branching_code_new_vals_init)
    for x in range(0, len(branching_code_old_vals_init)):
        branching_code_old_vals[x] = eval(branching_code_old_vals_init[x])
        branching_code_new_vals[x] = eval(branching_code_new_vals_init[x])
    
    # Check if the experiment exists.
    experiment = adaptlibs.GetDataAccessObject().experimentsSelectAllWhere(expid=expid)
    if len(experiment) == 0:
        experiments = adaptlibs.GetDataAccessObject().experimentsSelectAllAll()
        exp_list = []
        for item in experiments:
            exp_list.append(item[0])
        print('Experiment %s does not exist in the database. Currently-existing experiments: %s.' % (str(expid), exp_list))
        return False
    
    # Get all branches.
    all_branches = adaptlibs.GetDataAccessObject().branchesSelectAllWhere(expid=expid)
    
    # Get each job where a branching condition occurred as well as the code
    jobs_branching = adaptlibs.GetDataAccessObject().job_output_infoSelectInfoWhereKAndExpID(k='code', expid=expid)
    if len(jobs_branching) == 0:
        print('There do not appear to be any jobs that ended in a branching condition for experiment %s.' % str(expid))
        return False
    
    branching_jobs_to_branches = {}
    branching_branches_to_jobs = {}
    for job in jobs_branching:
        row = adaptlibs.GetDataAccessObject().jobsSelectAllWhereJobID(jobid=job[0])[0]
        branching_jobs_to_branches[row[0]] = {}
        branching_branches_to_jobs[row[1]] = {}
        branching_jobs_to_branches[row[0]]['branch'] = row[1]
        branching_branches_to_jobs[row[1]]['job'] = row[0]
        branching_jobs_to_branches[row[0]]['code'] = job[2]
        branching_branches_to_jobs[row[1]]['code'] = job[2]
    
    # Get all immediate children
    immediate_children = {}
    for parent in branching_branches_to_jobs:
        immediate_children[parent] = adaptlibs.children_identify(parent)
    
    all_branches_dict = {}
    for ind in all_branches:
        all_branches_dict[ind[0]] = ind
    
    # Tie each immediate child to a branching code and branch description ('sb 1', 'sb 2', ...)
    all_child_descs = {}
    for parent in immediate_children:
        for child in immediate_children[parent]:
            all_child_descs[child] = {}
            all_child_descs[child]['code'] = branching_branches_to_jobs[parent]['code']
            all_child_descs[child]['desc'] = all_branches_dict[child][3]
    
    # Get heritage for each branch
    (expid, root_branch, end_branches, heritage_all) = adaptlibs.get_exp_branch_heritage(expid)
    
    # Made a list of branches from the chosen branching_code.
    affected_immediate_children = []
    for child in all_child_descs:
        if all_child_descs[child]['code'] == str(branching_code):
            affected_immediate_children.append(child)
    
    # Check if there were any occurrences of the input branching code.
    if len(affected_immediate_children) == 0:
        print('There do not appear to be any occurrences of branching code %s in experiment %s.' % (str(branching_code), str(expid)))
        return False
    
    # Loop over all branches, and determine if any of the affected_immediate_children are in the heritage.
    all_branches_eventually_affected = {}
    for branch in all_branches_dict:
        for heritage in heritage_all[branch]:
            if heritage in affected_immediate_children:
                all_branches_eventually_affected[branch] = {}
                all_branches_eventually_affected[branch]['code'] = all_child_descs[heritage]['code']
                all_branches_eventually_affected[branch]['desc'] = all_child_descs[heritage]['desc']
                if progressive is True:
                    all_branches_eventually_affected[branch]['origprob'] = new_total_prob[branch]
                else:
                    all_branches_eventually_affected[branch]['origprob'] = all_branches_dict[branch][6]
    
    # All of the branches in all_branches_eventually_affected get their probabilities divided by the old conditional probability and multiplied by the new conditional probability associated with their description.
    old_cond_prob = {}
    new_cond_prob = {}
    new_total_prob = {}
    for branch in all_branches_eventually_affected:
        for x in range(0, len(branching_code_descs)):
            if branching_code_descs[x] == all_branches_eventually_affected[branch]['desc']:
                old_cond_prob[branch] = branching_code_old_vals[x]
                new_cond_prob[branch] = branching_code_new_vals[x]
                new_total_prob[branch] = (all_branches_eventually_affected[branch]['origprob'] / old_cond_prob[branch]) * new_cond_prob[branch]
    
    if commit is True:
        for branch in new_total_prob:
            adaptlibs.GetDataAccessObject().branchesUpdateProbabilityWhereBranchID(branchid=branch, probability=('%.8e' % new_total_prob[branch]))
    
    return (new_total_prob, all_branches_eventually_affected)

