#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ##########################
# ADAPT copyright statement:
# Copyright 2006 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software. This software is distributed under the GNU Lesser General Public License version 3 or later.
# ##########################
import threading
import adaptlibs
import time
import traceback
import os
import sys
import subprocess
import pickle
import tempfile
from misc_const import debug
from adapt_config import AdaptConfig
from adapt_logging import AdaptLogging


class job_starter_thread(threading.Thread):
    
    def __init__(self,
                 executor,
                 branchid,
                 jobscript,
                 experimentid,
                 parentid,
                 branchdescription,
                 terminate_early,
                 resuming_checkpoint,
                 needed_id,
                 needed_host,
                 needed_rundir,
                 probability,
                 probability_threshold_multiplier,
                 probability_threshold,
                 executable, threadDAO, scp_method):
        
        threading.Thread.__init__(self)
        self.MyThreadDataAccessObject = threadDAO
        self.executor = executor
        self.branchid = branchid
        self.jobscript = jobscript
        self.experimentid = experimentid
        self.parentid = parentid
        self.branchdescription = branchdescription
        self.terminate_early = terminate_early
        self.resuming_checkpoint = resuming_checkpoint
        self.needed_id = needed_id
        self.needed_host = needed_host
        self.needed_rundir = needed_rundir
        self.probability = probability
        self.probability_threshold_multiplier = probability_threshold_multiplier
        self.probability_threshold = probability_threshold
        self.executable = executable
        self.scp_method = scp_method
    
    def run(self):
        executor = self.executor
        branchid = self.branchid
        jobscript = self.jobscript
        experimentid = self.experimentid
        parentid = self.parentid
        branchdescription = self.branchdescription
        terminate_early = self.terminate_early
        resuming_checkpoint = self.resuming_checkpoint
        needed_id = self.needed_id
        needed_host = self.needed_host
        needed_rundir = self.needed_rundir
        probability = self.probability
        probability_threshold_multiplier = self.probability_threshold_multiplier
        probability_threshold = self.probability_threshold
        executable = self.executable
        
        if resuming_checkpoint: # resuming a checkpointed branch
            jobid = needed_id
            self.MyThreadDataAccessObject.jobsUpdateStateWhereID(state='R', jobid=jobid)
            if AdaptConfig.get_ssh_shared_dirs():
                self.MyThreadDataAccessObject.jobsUpdateHostnameWhereID(hostname=executor.get_hostname(), jobid=jobid)
        else:
            self.MyThreadDataAccessObject.jobsInsertAllAll(branchid, executor.get_hostname())
            jobid = self.MyThreadDataAccessObject.dualSelectLastInsertID()[0][0]
        
        try:
            tries = 0
            while tries < 5:
                try:
                    remote_container_store = executor.make_adaptemp_container()
                    (tempfd, tempfn) = tempfile.mkstemp(dir=AdaptConfig.get_server_adaptemp(), suffix='.scr')
                    os.close(tempfd)
                    tries = 5
                except:
                    print('%s: Failed to create new adaptemp container for branchid %s and jobid %s. Retrying.' % (str(time.asctime()), str(branchid), str(jobid)))
                    tries += 1
                    time.sleep(10)
            
            if 'tempfn' not in vars():
                print('%s: Failed 5 times to create new adaptemp container for branchid %s and jobid %s. Failing.' % (str(time.asctime()), str(branchid), str(jobid)))
                raise Exception('%s: Failed 5 times to create new adaptemp container for branchid %s and jobid %s. Failing job creation thread.' % (str(time.asctime()), str(branchid), str(jobid)))
            if int(sys.version[0]) > 2:
                jobscript = bytes(jobscript, 'utf-8')
            open(tempfn, 'wb').write(jobscript)
            
            DatabaseIOFile = os.path.join(os.path.dirname(sys.argv[0]), 'AdaptDatabaseIO.py')
            DbConnectionFile = os.path.join(os.path.dirname(sys.argv[0]), 'DbConnection.py')
            
            libsfile = os.path.join(os.path.dirname(sys.argv[0]), 'adaptlibs.py')
            supervisorfile = os.path.join(os.path.dirname(sys.argv[0]), 'adapt-supervisor')
            misc_const_file = os.path.join(os.path.dirname(sys.argv[0]), "misc_const.py")
            adapt_config_file = os.path.join(os.path.dirname(sys.argv[0]), "adapt_config.py")
            tries = 0
            copy_files_successful = False
            while tries < 5:
                try:
                    executor.copy_files_to_adaptemp_container([tempfn, libsfile, supervisorfile, misc_const_file, DatabaseIOFile, DbConnectionFile, adapt_config_file], remote_container_store, make_executable=True)
                    tries = 5
                    copy_files_successful = True
                except:
                    print('%s: Failed to copy files to adaptemp container for branchid %s and jobid %s. Retrying.' % (str(time.asctime()), str(branchid), str(jobid)))
                    tries += 1
                    time.sleep(10)
                    
            if copy_files_successful is False:
                    print('%s: Failed 5 times to to copy files to adaptemp container for branchid %s and jobid %s. Retrying.' % (str(time.asctime()), str(branchid), str(jobid)))
                    AdaptLogging.log('copy_files_to_adaptemp_container', 'An error has occurred while executing copy_files_to_adaptemp_container. 5 retries failed, canceling job %s.' % (jobid), logfile=AdaptConfig.get_server_log(), force=True)
                    sys.exit(1)
            os.remove(tempfn)
            
            jobdir = os.path.join(executor.directory, 'e%d' % (experimentid), 'b%06d-j%06d' % (branchid, jobid))
            if debug:
                print('ADAPT job execution directory is %s.' % jobdir)
            
            if not resuming_checkpoint:
                self.MyThreadDataAccessObject.jobsUpdateRundirWhereID(jobdir=jobdir, jobid=jobid)
                executor.cleanup_remote_container(jobdir)
                tries = 0
                while tries < 5:
                    try:
                        executor.make_remote_container(jobdir)
                        tries = 5
                    except:
                        print('%s: Failed to create new adaptemp container for branchid %s and jobid %s on host %s. Retrying.' % (str(time.asctime()), str(branchid), str(jobid), str(executor.get_hostname())))
                        tries += 1
                        time.sleep(10)
                
                # grab input blobs
                local_blobdir = tempfile.mkdtemp(dir=AdaptConfig.get_server_adaptemp())
                recs = self.MyThreadDataAccessObject.branch_input_blobsSelectAllWhere(branchid=branchid)
                for rec in recs:
                    fn = rec[1]
                    permissions = rec[2]
                    dat = rec[3]
                    locfn = os.path.join(local_blobdir, fn)
                    open(locfn, 'wb').write(dat)
                    os.chmod(locfn, permissions)
                    executor.copy_files_to_remote_container(locfn, jobdir)
                    os.remove(locfn)
                recs = None
                os.rmdir(local_blobdir)
                
                # grab input refs
                recs = self.MyThreadDataAccessObject.branch_input_refsSelectAllWhere(branchid=branchid)
                
                # If using 2stepcopy or shared_dirs use the old way.
                if AdaptConfig.get_ssh_2stepcopy() or AdaptConfig.get_ssh_shared_dirs():
                    for rec in recs:
                        hn = rec[1]
                        fnhere = rec[2]
                        fnthere = rec[3]
                        
                        if AdaptConfig.get_ssh_2stepcopy():
                            (tempfd, tempfn) = tempfile.mkstemp(dir=AdaptConfig.get_server_adaptemp())
                            os.close(tempfd)
                            cmd = '%s %s:%s %s' % (self.scp_method, hn, fnhere, tempfn)
                            cmd += ' && %s %s %s:%s' % (self.scp_method, tempfn, executor.get_hostname(), os.path.join(jobdir, fnthere))
                            cmd += ' && rm -f %s' % (tempfn)
                            #print('2stepcopy command is', cmd)
                        else:
                            cmd = '%s %s:%s %s:%s' % (self.scp_method, hn, fnhere, executor.get_hostname(), os.path.join(jobdir, fnthere))
                        if debug:
                            print('+%s' % (cmd))
                        
                        # Using a while loop to retry until successful or 5 times.
                        copy_success_final = False
                        copy_attempts = 0
                        while (copy_success_final is False) and (copy_attempts < 5):
                            before = time.time()
                            tries = 0
                            while tries < 5:
                                try:
                                    if AdaptConfig.get_ssh_shared_dirs():
                                        # If we use a shared file system, copy directly.
                                        try:
                                            import shutil
                                            shutil.copy2(fnhere, os.path.join(jobdir, fnthere))
                                            stderr = ''
                                        except Exception as ex:
                                            print('Error using shutil.copy2 to copy %s to %s. Error message: %s' % (fnhere, os.path.join(jobdir, fnthere), str(ex)))
                                            stderr = ('Error using shutil.copy2 to copy %s to %s.' % (fnhere, os.path.join(jobdir, fnthere)))
                                    else:
                                        p = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
                                        stdout, stderr = p.communicate()
                                    tries = 5
                                except:
                                    print('%s: Failed to execute %s. Retrying.' % (str(time.asctime()), cmd))
                                    tries += 1
                                    time.sleep(10)
                            if len(stderr) == 0:
                                copy_success = True
                            else:
                                copy_success = False
                            after = time.time()
                            msg = 'Time required for '
                            if hn == executor.get_hostname():
                                msg += 'local'
                            else:
                                msg += 'remote'
                            elapsed = after - before
                            msg += ' copy %s:%s to %s:%s is %f\n' % (hn, fnhere, executor.get_hostname(), os.path.join(jobdir, fnthere), elapsed)
                            if copy_success is True:
                                AdaptLogging.log('copy_timings', msg, logfile=AdaptConfig.get_server_log(), force=True)
                                copy_success_final = True
                                copy_attempts += 1
                            else:
                                AdaptLogging.log('copy_timings', 'An error has occurred while executing the command "%s". Error "%s" was returned.' % (cmd, stderr), logfile=AdaptConfig.get_server_log(), force=True)
                                copy_attempts += 1
                                time.sleep(10)
                        
                        if copy_attempts >= 5 and copy_success_final is False:
                            AdaptLogging.log('copy_timings', 'An error has occurred while executing the command "%s". Error "%s" was returned. 5 retries failed, canceling job %s.' % (cmd, stderr, jobid), logfile=AdaptConfig.get_server_log(), force=True)
                            sys.exit(1)
                else:
                    if len(recs) > 0:
                        sourcehost = recs[0][1]
                        sourcefiles = [x[2] for x in recs]
                        # Using a while loop to retry until successful or 5 times.
                        tries = 0
                        copy_files_successful = False
                        while tries < 5:
                            #try:
                            before = time.time()
                            (scp, sourcehost, file_str, hostname, destcontainer) = executor.copy_remote_files_to_remote_container(sourcehost, sourcefiles, executor.get_hostname(), jobdir, make_executable=False)
                            tries = 5
                            copy_files_successful = True
                            #except:
                            #    print('%s: Failed to copy files %s to remote container for branchid %s and jobid %s on host %s. Retrying.' % (str(time.asctime()), sourcefiles, str(branchid), str(jobid), str(executor.get_hostname())))
                            #    tries += 1
                            #    time.sleep(10)
                        
                        if copy_files_successful is False:
                                print('%s: An error has occurred while executing copy_remote_files_to_remote_container. 5 retries failed, canceling job %s.' % (time.asctime(), jobid))
                                AdaptLogging.log('copy_remote_files_to_remote_container', 'An error has occurred while executing copy_remote_files_to_remote_container. 5 retries failed, canceling job %s.' % (jobid), logfile=AdaptConfig.get_server_log(), force=True)
                                sys.exit(1)
                        after = time.time()
                        msg = 'Time required for remote'
                        elapsed = after - before
                        msg += ' copy %s:%s to %s:%s is %f\n' % (sourcehost, sourcefiles, executor.get_hostname(), jobdir, elapsed)
                        AdaptLogging.log('copy_timings', msg, logfile=AdaptConfig.get_server_log(), force=True)
            
            if parentid:
                root = '0'
            else:
                root = '1'
            
            try:
                resuming_checkpoint
            except NameError:
                resuming_checkpoint = None
            
            if resuming_checkpoint:
                resuming = '1'
                print("%s: Resuming branchid %s with jobid %s on host %s." % (str(time.asctime()), branchid, jobid, executor.get_hostname()))
            else:
                resuming = '0'
            
            remote_script = os.path.join(remote_container_store, os.path.basename(tempfn))
            
            # set up env for remote
            remote_env = {
                'NCENGINE_INTERNAL_BRANCHID': str(branchid),
                'NCENGINE_INTERNAL_JOBID': str(jobid),
                'NCENGINE_INTERNAL_EXPERIMENTID': str(experimentid),
                'NCENGINE_INTERNAL_DESCRIPTION': branchdescription,
                'NCENGINE_INTERNAL_EXPERIMENT_PROBABILITY_THRESHOLD': str(probability_threshold),
                'NCENGINE_INTERNAL_INHERITED_PROBABILITY': str(probability),
                'NCENGINE_INTERNAL_INHERITED_PROBABILITY_THRESHOLD_MULTIPLIER': str(probability_threshold_multiplier),
                'THISSCRIPT': remote_script,
                'NCENGINE_TERMINATE_EARLY': str(terminate_early),
                'NCENGINE_ROOT': root,
                'NCENGINE_RESUMING_CHECKPOINT': resuming,
                'NCENGINE_HOSTNAME': executor.get_hostname(),
                'NCENGINE_EXECUTABLE': executable}
            
            if os.getenv('NCENGINE_RCFILE'):
                remote_env['NCENGINE_RCFILE'] = os.getenv('NCENGINE_RCFILE')
            
            rows = self.MyThreadDataAccessObject.experiment_environment_variablesSelectAllWhere(experimentid)
            expid, k, v = rows[0]
            remote_env[k] = v
            tries = 0
            copy_files_successful = False
            while tries < 5:
                try:
                    (tempfd, picklefn) = tempfile.mkstemp(dir=remote_container_store, suffix='.pickle')
                    os.close(tempfd)
                    picklefn_base = os.path.basename(picklefn)
                    pickle.dump(remote_env, open(picklefn, 'wb'), protocol=0)
                    tries = 5
                    copy_files_successful = True
                except:
                    print('%s: Failed to create or copy environment variables to remote container container for branchid %s and jobid %s on host %s. Retrying.' % (str(time.asctime()), str(branchid), str(jobid), str(executor.get_hostname())))
                    tries += 1
                    time.sleep(10)
            
            if copy_files_successful is False:
                    print('%s: Failed 5 times to create or copy environment variables to remote container container for branchid %s and jobid %s on host %s. Quitting.' % (str(time.asctime()), str(branchid), str(jobid), str(executor.get_hostname())))
                    AdaptLogging.log('copy_files_to_remote_container', 'An error has occurred while executing copy_files_to_remote_container. 5 retries failed, canceling job %s.' % (jobid), logfile=AdaptConfig.get_server_log(), force=True)
                    sys.exit(1)
                
            rovestr = ''
            if AdaptConfig.get_rove_hostname() and AdaptConfig.get_rove_port():
                rovestr = '-rove %s:%d' % (AdaptConfig.get_rove_hostname(), AdaptConfig.get_rove_port())
            
            remotecmd = 'cd %s && %s -env_pickled_dict %s %s adapt-stdout.txt adapt-stderr.txt %04d %s && cd / && rm -fr %s' % (jobdir, os.path.join(remote_container_store, 'adapt-supervisor'), os.path.join(remote_container_store, picklefn_base), rovestr, branchid, remote_script, remote_container_store)
            
            executor.job_start(remotecmd)
        except Exception:
            print(str(time.asctime()) + ': Caught an exception in ADAPT module clsJobStarterThread.py for branchid %s and jobid %s on host %s.' % (str(branchid), str(jobid), str(executor.get_hostname())))
            traceback.print_exc()
            self.MyThreadDataAccessObject.jobsUpdateStateWhereID(state='A', jobid=jobid)
            executor.job_mark_as_failed()
        finally:
            self.MyThreadDataAccessObject.ExplicitlyClose()
            del self.MyThreadDataAccessObject
